from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from . import session
from . import channel
from . import sftp
from . import exceptions
from . import sftp_handles
from . import tunnel
from . import utils
from . import libssh_enums as enums



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130


class OptionError(Exception):
    """Raised on errors getting/setting options"""


class BaseSSHError(Exception):
    """Base class for all errors produced by libssh"""


class SSHError(BaseSSHError):
    """Raised on errors returned by libssh.

    This is the general error libssh returns when any error
    occurs, whether that be server error or API error.

    Some functions like authentication have more specific error
    codes.
    """


class OtherError(BaseSSHError):
    """Raised on other non-specific fatal errors"""


class AuthenticationDenied(BaseSSHError):
    """Raised on authentication denied errors"""


class AuthenticationError(BaseSSHError):
    """Raised on fatal errors authenticating"""


class AuthenticationPartial(BaseSSHError):
    """Raised on partial authentication"""


class KeyExportError(BaseSSHError):
    """Raised on errors exporting key"""


class KeyImportError(BaseSSHError):
    """Raised on errors importing key"""


class KeyGenerationError(BaseSSHError):
    """Raised on errors generating key"""


class EOF(BaseSSHError):
    """Raised on EOF from remote channel"""


class InvalidAPIUse(BaseSSHError):
    """Raised on invalid uses of the API"""


class Disconnected(BaseSSHError):
    """Raised on disconnection errors"""


class UnImplemented(BaseSSHError):
    """Raised on unimplemented errors"""


class GSSAPIError(BaseSSHError):
    """Raised on GSS API errors"""


class GSSAPIErrorTok(BaseSSHError):
    """Raised on GSS API token errors"""


class RequestFailure(BaseSSHError):
    """Raised on SSH request failures"""


class ChannelOpenFailure(BaseSSHError):
    """Raised on SSH channel open failures"""


class HostNotAllowedToConnect(BaseSSHError):
    """Raised on host not allowed to connect errors"""


class ProtocolError(BaseSSHError):
    """Raised on protocol errors"""


class KeyExchangeFailed(BaseSSHError):
    """Raised on key exchange failures"""


class HostAuthenticationFailed(BaseSSHError):
    """Raised on host authentication failures"""


class MACError(BaseSSHError):
    """Raised on MAC errors"""


class CompressionError(BaseSSHError):
    """Raised on compression errors"""


class ServiceNotAvailable(BaseSSHError):
    """Raised on service not available errors"""


class ProtocolVersionNotSupport(BaseSSHError):
    """Raised on protocol version not supported"""


class HostKeyNotVerifiable(BaseSSHError):
    """Raised on host key not verifiable errors"""


class ConnectionLost(BaseSSHError):
    """Raised on connection lost"""


class TooManyConnections(BaseSSHError):
    """Raised on too many connection errors"""


class AdministrativelyProhibited(BaseSSHError):
    """Raised on administratively prohibited errors"""


class ConnectFailed(BaseSSHError):
    """Raised on connect failure"""


class UnknownChannelType(BaseSSHError):
    """Raised on unknown channel type"""


class ResourceShortage(BaseSSHError):
    """Raised on resource shortage errors"""


class SFTPError(BaseSSHError):
    """Raised on SFTP errors"""


class SFTPHandleError(SFTPError):
    """Raised on SFTP handle errors"""


class ChannelClosed(BaseSSHError):
    """Raised on operations on closed channels"""



# This file is part of redlibssh.
# Copyright (C) 2022 Red_M

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import enum

from . cimport c_sftp
from . cimport c_ssh
from . cimport c_ssh2
from . cimport c_callbacks

#ssh
class SSH(enum.Enum):
    MD5_DIGEST_LEN = c_ssh.MD5_DIGEST_LEN
    CRYPT = c_ssh.SSH_CRYPT
    MAC = c_ssh.SSH_MAC
    COMP = c_ssh.SSH_COMP
    LANG = c_ssh.SSH_LANG
    #Errors
    OK = c_ssh.SSH_OK
    ERROR = c_ssh.SSH_ERROR
    AGAIN = c_ssh.SSH_AGAIN
    EOF = c_ssh.SSH_EOF
    INVALID_SOCKET = c_ssh.SSH_INVALID_SOCKET
    #States
    CLOSED = c_ssh.SSH_CLOSED
    READ_PENDING = c_ssh.SSH_READ_PENDING
    CLOSED_ERROR = c_ssh.SSH_CLOSED_ERROR
    WRITE_PENDING = c_ssh.SSH_WRITE_PENDING

class Kex(enum.Enum):
    KEX = c_ssh.SSH_KEX
    HOSTKEYS = c_ssh.SSH_HOSTKEYS
    CRYPT_C_S = c_ssh.SSH_CRYPT_C_S
    CRYPT_S_C = c_ssh.SSH_CRYPT_S_C
    MAC_C_S = c_ssh.SSH_MAC_C_S
    MAC_S_C = c_ssh.SSH_MAC_S_C
    COMP_C_S = c_ssh.SSH_COMP_C_S
    COMP_S_C = c_ssh.SSH_COMP_S_C
    LANG_C_S = c_ssh.SSH_LANG_C_S
    LANG_S_C = c_ssh.SSH_LANG_S_C

class Auth(enum.Enum):
    SUCCESS = c_ssh.SSH_AUTH_SUCCESS
    DENIED = c_ssh.SSH_AUTH_DENIED
    PARTIAL = c_ssh.SSH_AUTH_PARTIAL
    INFO = c_ssh.SSH_AUTH_INFO
    AGAIN = c_ssh.SSH_AUTH_AGAIN
    ERROR = c_ssh.SSH_AUTH_ERROR

class Auth_Method(enum.Enum):
    UNKNOWN = c_ssh.SSH_AUTH_METHOD_UNKNOWN
    NONE = c_ssh.SSH_AUTH_METHOD_NONE
    PASSWORD = c_ssh.SSH_AUTH_METHOD_PASSWORD
    PUBLICKEY = c_ssh.SSH_AUTH_METHOD_PUBLICKEY
    HOSTBASED = c_ssh.SSH_AUTH_METHOD_HOSTBASED
    INTERACTIVE = c_ssh.SSH_AUTH_METHOD_INTERACTIVE
    GSSAPI_MIC = c_ssh.SSH_AUTH_METHOD_GSSAPI_MIC

    # enum ssh_requests_e:
class Request(enum.Enum):
    AUTH = c_ssh.SSH_REQUEST_AUTH
    CHANNEL_OPEN = c_ssh.SSH_REQUEST_CHANNEL_OPEN
    CHANNEL = c_ssh.SSH_REQUEST_CHANNEL
    SERVICE = c_ssh.SSH_REQUEST_SERVICE
    GLOBAL = c_ssh.SSH_REQUEST_GLOBAL

    # enum ssh_channel_type_e:
class Channel(enum.Enum):
    UNKNOWN = c_ssh.SSH_CHANNEL_UNKNOWN
    SESSION = c_ssh.SSH_CHANNEL_SESSION
    DIRECT_TCPIP = c_ssh.SSH_CHANNEL_DIRECT_TCPIP
    FORWARDED_TCPIP = c_ssh.SSH_CHANNEL_FORWARDED_TCPIP
    X11 = c_ssh.SSH_CHANNEL_X11
    AUTH_AGENT = c_ssh.SSH_CHANNEL_AUTH_AGENT

# enum ssh_channel_requests_e:
class Channel_Request(enum.Enum):
    UNKNOWN = c_ssh.SSH_CHANNEL_REQUEST_UNKNOWN
    PTY = c_ssh.SSH_CHANNEL_REQUEST_PTY
    EXEC = c_ssh.SSH_CHANNEL_REQUEST_EXEC
    SHELL = c_ssh.SSH_CHANNEL_REQUEST_SHELL
    ENV = c_ssh.SSH_CHANNEL_REQUEST_ENV
    SUBSYSTEM = c_ssh.SSH_CHANNEL_REQUEST_SUBSYSTEM
    WINDOW_CHANGE = c_ssh.SSH_CHANNEL_REQUEST_WINDOW_CHANGE
    X11 = c_ssh.SSH_CHANNEL_REQUEST_X11

    # enum ssh_global_requests_e:
class Global_Request(enum.Enum):
    UNKNOWN = c_ssh.SSH_GLOBAL_REQUEST_UNKNOWN
    TCPIP_FORWARD = c_ssh.SSH_GLOBAL_REQUEST_TCPIP_FORWARD
    CANCEL_TCPIP_FORWARD = c_ssh.SSH_GLOBAL_REQUEST_CANCEL_TCPIP_FORWARD
    KEEPALIVE = c_ssh.SSH_GLOBAL_REQUEST_KEEPALIVE

    # enum ssh_publickey_state_e:
class PublicKeyState(enum.Enum):
    ERROR = c_ssh.SSH_PUBLICKEY_STATE_ERROR
    NONE = c_ssh.SSH_PUBLICKEY_STATE_NONE
    VALID = c_ssh.SSH_PUBLICKEY_STATE_VALID
    WRONG = c_ssh.SSH_PUBLICKEY_STATE_WRONG

    # enum ssh_server_known_e:
class Server(enum.Enum):
    ERROR = c_ssh.SSH_SERVER_ERROR
    NOT_KNOWN = c_ssh.SSH_SERVER_NOT_KNOWN
    KNOWN_OK = c_ssh.SSH_SERVER_KNOWN_OK
    KNOWN_CHANGED = c_ssh.SSH_SERVER_KNOWN_CHANGED
    FOUND_OTHER = c_ssh.SSH_SERVER_FOUND_OTHER
    FILE_NOT_FOUND = c_ssh.SSH_SERVER_FILE_NOT_FOUND

    # Errors
    # enum ssh_error_types_e:
class ErrorTypes(enum.Enum):
    NO_ERROR = c_ssh.SSH_NO_ERROR
    REQUEST_DENIED = c_ssh.SSH_REQUEST_DENIED
    FATAL = c_ssh.SSH_FATAL
    EINTR = c_ssh.SSH_EINTR

    # enum ssh_keytypes_e:
class KeyType(enum.Enum):
    UNKNOWN = c_ssh.SSH_KEYTYPE_UNKNOWN
    DSS = c_ssh.SSH_KEYTYPE_DSS
    RSA = c_ssh.SSH_KEYTYPE_RSA
    RSA1 = c_ssh.SSH_KEYTYPE_RSA1
    ECDSA = c_ssh.SSH_KEYTYPE_ECDSA  # Deprecated
    ED25519 = c_ssh.SSH_KEYTYPE_ED25519
    DSS_CERT01 = c_ssh.SSH_KEYTYPE_DSS_CERT01
    RSA_CERT01 = c_ssh.SSH_KEYTYPE_RSA_CERT01
    ECDSA_P256 = c_ssh.SSH_KEYTYPE_ECDSA_P256
    ECDSA_P384 = c_ssh.SSH_KEYTYPE_ECDSA_P384
    ECDSA_P521 = c_ssh.SSH_KEYTYPE_ECDSA_P521
    ECDSA_P256_CERT01 = c_ssh.SSH_KEYTYPE_ECDSA_P256_CERT01
    ECDSA_P384_CERT01 = c_ssh.SSH_KEYTYPE_ECDSA_P384_CERT01
    ECDSA_P521_CERT01 = c_ssh.SSH_KEYTYPE_ECDSA_P521_CERT01
    ED25519_CERT01 = c_ssh.SSH_KEYTYPE_ED25519_CERT01

    # enum ssh_keycmp_e:
class KeyCMP(enum.Enum):
    PUBLIC = c_ssh.SSH_KEY_CMP_PUBLIC
    PRIVATE = c_ssh.SSH_KEY_CMP_PRIVATE

class Log(enum.Enum):
    NOLOG = c_ssh.SSH_LOG_NOLOG
    PROTOCOL = c_ssh.SSH_LOG_PROTOCOL
    PACKET = c_ssh.SSH_LOG_PACKET
    FUNCTIONS = c_ssh.SSH_LOG_FUNCTIONS
    RARE = c_ssh.SSH_LOG_RARE
    NONE = c_ssh.SSH_LOG_NONE
    WARN = c_ssh.SSH_LOG_WARN
    INFO = c_ssh.SSH_LOG_INFO
    DEBUG = c_ssh.SSH_LOG_DEBUG
    TRACE = c_ssh.SSH_LOG_TRACE

    # enum ssh_options_e:
class Options(enum.Enum):
    HOST = c_ssh.SSH_OPTIONS_HOST
    PORT = c_ssh.SSH_OPTIONS_PORT
    PORT_STR = c_ssh.SSH_OPTIONS_PORT_STR
    FD = c_ssh.SSH_OPTIONS_FD
    USER = c_ssh.SSH_OPTIONS_USER
    SSH_DIR = c_ssh.SSH_OPTIONS_SSH_DIR
    IDENTITY = c_ssh.SSH_OPTIONS_IDENTITY
    ADD_IDENTITY = c_ssh.SSH_OPTIONS_ADD_IDENTITY
    KNOWNHOSTS = c_ssh.SSH_OPTIONS_KNOWNHOSTS
    TIMEOUT = c_ssh.SSH_OPTIONS_TIMEOUT
    TIMEOUT_USEC = c_ssh.SSH_OPTIONS_TIMEOUT_USEC
    SSH1 = c_ssh.SSH_OPTIONS_SSH1
    SSH2 = c_ssh.SSH_OPTIONS_SSH2
    LOG_VERBOSITY = c_ssh.SSH_OPTIONS_LOG_VERBOSITY
    LOG_VERBOSITY_STR = c_ssh.SSH_OPTIONS_LOG_VERBOSITY_STR
    CIPHERS_C_S = c_ssh.SSH_OPTIONS_CIPHERS_C_S
    CIPHERS_S_C = c_ssh.SSH_OPTIONS_CIPHERS_S_C
    COMPRESSION_C_S = c_ssh.SSH_OPTIONS_COMPRESSION_C_S
    COMPRESSION_S_C = c_ssh.SSH_OPTIONS_COMPRESSION_S_C
    PROXYCOMMAND = c_ssh.SSH_OPTIONS_PROXYCOMMAND
    BINDADDR = c_ssh.SSH_OPTIONS_BINDADDR
    STRICTHOSTKEYCHECK = c_ssh.SSH_OPTIONS_STRICTHOSTKEYCHECK
    COMPRESSION = c_ssh.SSH_OPTIONS_COMPRESSION
    COMPRESSION_LEVEL = c_ssh.SSH_OPTIONS_COMPRESSION_LEVEL
    KEY_EXCHANGE = c_ssh.SSH_OPTIONS_KEY_EXCHANGE
    HOSTKEYS = c_ssh.SSH_OPTIONS_HOSTKEYS
    GSSAPI_SERVER_IDENTITY = c_ssh.SSH_OPTIONS_GSSAPI_SERVER_IDENTITY
    GSSAPI_CLIENT_IDENTITY = c_ssh.SSH_OPTIONS_GSSAPI_CLIENT_IDENTITY
    GSSAPI_DELEGATE_CREDENTIALS = c_ssh.SSH_OPTIONS_GSSAPI_DELEGATE_CREDENTIALS
    HMAC_C_S = c_ssh.SSH_OPTIONS_HMAC_C_S
    HMAC_S_C = c_ssh.SSH_OPTIONS_HMAC_S_C
    PASSWORD_AUTH = c_ssh.SSH_OPTIONS_PASSWORD_AUTH
    PUBKEY_AUTH = c_ssh.SSH_OPTIONS_PUBKEY_AUTH
    KBDINT_AUTH = c_ssh.SSH_OPTIONS_KBDINT_AUTH
    GSSAPI_AUTH = c_ssh.SSH_OPTIONS_GSSAPI_AUTH
    GLOBAL_KNOWNHOSTS = c_ssh.SSH_OPTIONS_GLOBAL_KNOWNHOSTS
    NODELAY = c_ssh.SSH_OPTIONS_NODELAY

class SCP(enum.Enum):
    WRITE = c_ssh.SSH_SCP_WRITE
    READ = c_ssh.SSH_SCP_READ
    RECURSIVE = c_ssh.SSH_SCP_RECURSIVE
        # enum ssh_scp_request_types:
class SCP_Request(enum.Enum):
    NEWDIR = c_ssh.SSH_SCP_REQUEST_NEWDIR
    NEWFILE = c_ssh.SSH_SCP_REQUEST_NEWFILE
    EOF = c_ssh.SSH_SCP_REQUEST_EOF
    ENDDIR = c_ssh.SSH_SCP_REQUEST_ENDDIR
    WARNING = c_ssh.SSH_SCP_REQUEST_WARNING

    # enum ssh_connector_flags_e:
class Connector(enum.Enum):
    STDOUT = c_ssh.SSH_CONNECTOR_STDOUT
    STDERR = c_ssh.SSH_CONNECTOR_STDERR
    BOTH = c_ssh.SSH_CONNECTOR_BOTH
#callbacks
class Callbacks_Socket(enum.Enum):
    FLOW_WRITEWILLBLOCK = c_callbacks.SSH_SOCKET_FLOW_WRITEWILLBLOCK
    FLOW_WRITEWONTBLOCK = c_callbacks.SSH_SOCKET_FLOW_WRITEWONTBLOCK
    EXCEPTION_EOF = c_callbacks.SSH_SOCKET_EXCEPTION_EOF
    EXCEPTION_ERROR = c_callbacks.SSH_SOCKET_EXCEPTION_ERROR
    CONNECTED_OK = c_callbacks.SSH_SOCKET_CONNECTED_OK
    CONNECTED_ERROR = c_callbacks.SSH_SOCKET_CONNECTED_ERROR
    CONNECTED_TIMEOUT = c_callbacks.SSH_SOCKET_CONNECTED_TIMEOUT

#ssh2
class SSH2_Msg(enum.Enum):
    DISCONNECT = c_ssh2.SSH2_MSG_DISCONNECT
    IGNORE = c_ssh2.SSH2_MSG_IGNORE
    UNIMPLEMENTED = c_ssh2.SSH2_MSG_UNIMPLEMENTED
    DEBUG = c_ssh2.SSH2_MSG_DEBUG
    SERVICE_REQUEST = c_ssh2.SSH2_MSG_SERVICE_REQUEST
    SERVICE_ACCEPT = c_ssh2.SSH2_MSG_SERVICE_ACCEPT
    KEXINIT = c_ssh2.SSH2_MSG_KEXINIT
    NEWKEYS = c_ssh2.SSH2_MSG_NEWKEYS
    ECMQV_INIT = c_ssh2.SSH2_MSG_ECMQV_INIT
    ECMQV_REPLY = c_ssh2.SSH2_MSG_ECMQV_REPLY
    KEXDH_INIT = c_ssh2.SSH2_MSG_KEXDH_INIT
    KEXDH_REPLY = c_ssh2.SSH2_MSG_KEXDH_REPLY
class SSH2_Msg_Kex(enum.Enum):
    ECDH_INIT = c_ssh2.SSH2_MSG_KEX_ECDH_INIT
    ECDH_REPLY = c_ssh2.SSH2_MSG_KEX_ECDH_REPLY
    DH_GEX_REQUEST_OLD = c_ssh2.SSH2_MSG_KEX_DH_GEX_REQUEST_OLD
    DH_GEX_GROUP = c_ssh2.SSH2_MSG_KEX_DH_GEX_GROUP
    DH_GEX_INIT = c_ssh2.SSH2_MSG_KEX_DH_GEX_INIT
    DH_GEX_REPLY = c_ssh2.SSH2_MSG_KEX_DH_GEX_REPLY
    DH_GEX_REQUEST = c_ssh2.SSH2_MSG_KEX_DH_GEX_REQUEST
class SSH2_Msg_UserAuth(enum.Enum):
    REQUEST = c_ssh2.SSH2_MSG_USERAUTH_REQUEST
    FAILURE = c_ssh2.SSH2_MSG_USERAUTH_FAILURE
    SUCCESS = c_ssh2.SSH2_MSG_USERAUTH_SUCCESS
    BANNER = c_ssh2.SSH2_MSG_USERAUTH_BANNER
    PK_OK = c_ssh2.SSH2_MSG_USERAUTH_PK_OK
    PASSWD_CHANGEREQ = c_ssh2.SSH2_MSG_USERAUTH_PASSWD_CHANGEREQ
    INFO_REQUEST = c_ssh2.SSH2_MSG_USERAUTH_INFO_REQUEST
    GSSAPI_RESPONSE = c_ssh2.SSH2_MSG_USERAUTH_GSSAPI_RESPONSE
    INFO_RESPONSE = c_ssh2.SSH2_MSG_USERAUTH_INFO_RESPONSE
    GSSAPI_TOKEN = c_ssh2.SSH2_MSG_USERAUTH_GSSAPI_TOKEN
    GSSAPI_EXCHANGE_COMPLETE = c_ssh2.SSH2_MSG_USERAUTH_GSSAPI_EXCHANGE_COMPLETE
    GSSAPI_ERROR = c_ssh2.SSH2_MSG_USERAUTH_GSSAPI_ERROR
    GSSAPI_ERRTOK = c_ssh2.SSH2_MSG_USERAUTH_GSSAPI_ERRTOK
    GSSAPI_MIC = c_ssh2.SSH2_MSG_USERAUTH_GSSAPI_MIC
class SSH2_Msg_Global(enum.Enum):
    REQUEST = c_ssh2.SSH2_MSG_GLOBAL_REQUEST
class SSH2_Msg_Request(enum.Enum):
    SUCCESS = c_ssh2.SSH2_MSG_REQUEST_SUCCESS
    FAILURE = c_ssh2.SSH2_MSG_REQUEST_FAILURE
class SSH2_Msg_Channel(enum.Enum):
    OPEN = c_ssh2.SSH2_MSG_CHANNEL_OPEN
    OPEN_CONFIRMATION = c_ssh2.SSH2_MSG_CHANNEL_OPEN_CONFIRMATION
    OPEN_FAILURE = c_ssh2.SSH2_MSG_CHANNEL_OPEN_FAILURE
    WINDOW_ADJUST = c_ssh2.SSH2_MSG_CHANNEL_WINDOW_ADJUST
    DATA = c_ssh2.SSH2_MSG_CHANNEL_DATA
    EXTENDED_DATA = c_ssh2.SSH2_MSG_CHANNEL_EXTENDED_DATA
    EOF = c_ssh2.SSH2_MSG_CHANNEL_EOF
    CLOSE = c_ssh2.SSH2_MSG_CHANNEL_CLOSE
    REQUEST = c_ssh2.SSH2_MSG_CHANNEL_REQUEST
    SUCCESS = c_ssh2.SSH2_MSG_CHANNEL_SUCCESS
    FAILURE = c_ssh2.SSH2_MSG_CHANNEL_FAILURE
class SSH2_Disconnect(enum.Enum):
    HOST_NOT_ALLOWED_TO_CONNECT = c_ssh2.SSH2_DISCONNECT_HOST_NOT_ALLOWED_TO_CONNECT
    PROTOCOL_ERROR = c_ssh2.SSH2_DISCONNECT_PROTOCOL_ERROR
    KEY_EXCHANGE_FAILED = c_ssh2.SSH2_DISCONNECT_KEY_EXCHANGE_FAILED
    HOST_AUTHENTICATION_FAILED = c_ssh2.SSH2_DISCONNECT_HOST_AUTHENTICATION_FAILED
    RESERVED = c_ssh2.SSH2_DISCONNECT_RESERVED
    MAC_ERROR = c_ssh2.SSH2_DISCONNECT_MAC_ERROR
    COMPRESSION_ERROR = c_ssh2.SSH2_DISCONNECT_COMPRESSION_ERROR
    SERVICE_NOT_AVAILABLE = c_ssh2.SSH2_DISCONNECT_SERVICE_NOT_AVAILABLE
    PROTOCOL_VERSION_NOT_SUPPORTED = c_ssh2.SSH2_DISCONNECT_PROTOCOL_VERSION_NOT_SUPPORTED
    HOST_KEY_NOT_VERIFIABLE = c_ssh2.SSH2_DISCONNECT_HOST_KEY_NOT_VERIFIABLE
    CONNECTION_LOST = c_ssh2.SSH2_DISCONNECT_CONNECTION_LOST
    BY_APPLICATION = c_ssh2.SSH2_DISCONNECT_BY_APPLICATION
    TOO_MANY_CONNECTIONS = c_ssh2.SSH2_DISCONNECT_TOO_MANY_CONNECTIONS
    AUTH_CANCELLED_BY_USER = c_ssh2.SSH2_DISCONNECT_AUTH_CANCELLED_BY_USER
    NO_MORE_AUTH_METHODS_AVAILABLE = c_ssh2.SSH2_DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE
    ILLEGAL_USER_NAME = c_ssh2.SSH2_DISCONNECT_ILLEGAL_USER_NAME
class SSH2_Open(enum.Enum):
    ADMINISTRATIVELY_PROHIBITED = c_ssh2.SSH2_OPEN_ADMINISTRATIVELY_PROHIBITED
    CONNECT_FAILED = c_ssh2.SSH2_OPEN_CONNECT_FAILED
    UNKNOWN_CHANNEL_TYPE = c_ssh2.SSH2_OPEN_UNKNOWN_CHANNEL_TYPE
    RESOURCE_SHORTAGE = c_ssh2.SSH2_OPEN_RESOURCE_SHORTAGE
class SSH2_ExtendedData(enum.Enum):
    STDERR = c_ssh2.SSH2_EXTENDED_DATA_STDERR

#sftp
class SFTP(enum.Enum):
    OPEN = c_sftp.SFTP_OPEN
    CLOSE = c_sftp.SFTP_CLOSE
    READ = c_sftp.SFTP_READ
    WRITE = c_sftp.SFTP_WRITE
    LSTAT = c_sftp.SFTP_LSTAT
    FSTAT = c_sftp.SFTP_FSTAT
    SETSTAT = c_sftp.SFTP_SETSTAT
    FSETSTAT = c_sftp.SFTP_FSETSTAT
    OPENDIR = c_sftp.SFTP_OPENDIR
    READDIR = c_sftp.SFTP_READDIR
    REMOVE = c_sftp.SFTP_REMOVE
    MKDIR = c_sftp.SFTP_MKDIR
    RMDIR = c_sftp.SFTP_RMDIR
    REALPATH = c_sftp.SFTP_REALPATH
    STAT = c_sftp.SFTP_STAT
    RENAME = c_sftp.SFTP_RENAME
    READLINK = c_sftp.SFTP_READLINK
    SYMLINK = c_sftp.SFTP_SYMLINK
class SFTP_FXP(enum.Enum):
    INIT = c_sftp.SSH_FXP_INIT
    VERSION = c_sftp.SSH_FXP_VERSION
    OPEN = c_sftp.SSH_FXP_OPEN
    CLOSE = c_sftp.SSH_FXP_CLOSE
    READ = c_sftp.SSH_FXP_READ
    WRITE = c_sftp.SSH_FXP_WRITE
    LSTAT = c_sftp.SSH_FXP_LSTAT
    FSTAT = c_sftp.SSH_FXP_FSTAT
    SETSTAT = c_sftp.SSH_FXP_SETSTAT
    FSETSTAT = c_sftp.SSH_FXP_FSETSTAT
    OPENDIR = c_sftp.SSH_FXP_OPENDIR
    READDIR = c_sftp.SSH_FXP_READDIR
    REMOVE = c_sftp.SSH_FXP_REMOVE
    MKDIR = c_sftp.SSH_FXP_MKDIR
    RMDIR = c_sftp.SSH_FXP_RMDIR
    REALPATH = c_sftp.SSH_FXP_REALPATH
    STAT = c_sftp.SSH_FXP_STAT
    RENAME = c_sftp.SSH_FXP_RENAME
    READLINK = c_sftp.SSH_FXP_READLINK
    SYMLINK = c_sftp.SSH_FXP_SYMLINK
    STATUS = c_sftp.SSH_FXP_STATUS
    HANDLE = c_sftp.SSH_FXP_HANDLE
    DATA = c_sftp.SSH_FXP_DATA
    NAME = c_sftp.SSH_FXP_NAME
    ATTRS = c_sftp.SSH_FXP_ATTRS
    EXTENDED = c_sftp.SSH_FXP_EXTENDED
    EXTENDED_REPLY = c_sftp.SSH_FXP_EXTENDED_REPLY
class SFTP_FileXfer_Attr(enum.Enum):
    SIZE = c_sftp.SSH_FILEXFER_ATTR_SIZE
    PERMISSIONS = c_sftp.SSH_FILEXFER_ATTR_PERMISSIONS
    ACCESSTIME = c_sftp.SSH_FILEXFER_ATTR_ACCESSTIME
    ACMODTIME = c_sftp.SSH_FILEXFER_ATTR_ACMODTIME
    CREATETIME = c_sftp.SSH_FILEXFER_ATTR_CREATETIME
    MODIFYTIME = c_sftp.SSH_FILEXFER_ATTR_MODIFYTIME
    ACL = c_sftp.SSH_FILEXFER_ATTR_ACL
    OWNERGROUP = c_sftp.SSH_FILEXFER_ATTR_OWNERGROUP
    SUBSECOND_TIMES = c_sftp.SSH_FILEXFER_ATTR_SUBSECOND_TIMES
    EXTENDED = c_sftp.SSH_FILEXFER_ATTR_EXTENDED
    UIDGID = c_sftp.SSH_FILEXFER_ATTR_UIDGID
# Types
class SFTP_FileXfer_Type(enum.Enum):
    REGULAR = c_sftp.SSH_FILEXFER_TYPE_REGULAR
    DIRECTORY = c_sftp.SSH_FILEXFER_TYPE_DIRECTORY
    SYMLINK = c_sftp.SSH_FILEXFER_TYPE_SYMLINK
    SPECIAL = c_sftp.SSH_FILEXFER_TYPE_SPECIAL
    UNKNOWN = c_sftp.SSH_FILEXFER_TYPE_UNKNOWN
class SFTP_FX(enum.Enum):
    OK = c_sftp.SSH_FX_OK
    EOF = c_sftp.SSH_FX_EOF
    NO_SUCH_FILE = c_sftp.SSH_FX_NO_SUCH_FILE
    PERMISSION_DENIED = c_sftp.SSH_FX_PERMISSION_DENIED
    FAILURE = c_sftp.SSH_FX_FAILURE
    BAD_MESSAGE = c_sftp.SSH_FX_BAD_MESSAGE
    NO_CONNECTION = c_sftp.SSH_FX_NO_CONNECTION
    CONNECTION_LOST = c_sftp.SSH_FX_CONNECTION_LOST
    OP_UNSUPPORTED = c_sftp.SSH_FX_OP_UNSUPPORTED
    INVALID_HANDLE = c_sftp.SSH_FX_INVALID_HANDLE
    NO_SUCH_PATH = c_sftp.SSH_FX_NO_SUCH_PATH
    FILE_ALREADY_EXISTS = c_sftp.SSH_FX_FILE_ALREADY_EXISTS
    WRITE_PROTECT = c_sftp.SSH_FX_WRITE_PROTECT
    NO_MEDIA = c_sftp.SSH_FX_NO_MEDIA
class SFTP_FXF(enum.Enum):
    READ = c_sftp.SSH_FXF_READ
    WRITE = c_sftp.SSH_FXF_WRITE
    APPEND = c_sftp.SSH_FXF_APPEND
    CREAT = c_sftp.SSH_FXF_CREAT
    TRUNC = c_sftp.SSH_FXF_TRUNC
    EXCL = c_sftp.SSH_FXF_EXCL
    TEXT = c_sftp.SSH_FXF_TEXT
class SFTP_FXF_Rename(enum.Enum):
    OVERWRITE = c_sftp.SSH_FXF_RENAME_OVERWRITE
    ATOMIC = c_sftp.SSH_FXF_RENAME_ATOMIC
    NATIVE = c_sftp.SSH_FXF_RENAME_NATIVE
class SFTP_S(enum.Enum):
    IFMT = c_sftp.SSH_S_IFMT
    IFSOCK = c_sftp.SSH_S_IFSOCK
    IFLNK = c_sftp.SSH_S_IFLNK
    IFREG = c_sftp.SSH_S_IFREG
    IFBLK = c_sftp.SSH_S_IFBLK
    IFDIR = c_sftp.SSH_S_IFDIR
    IFCHR = c_sftp.SSH_S_IFCHR
    IFIFO = c_sftp.SSH_S_IFIFO
class SFTP_FXE(enum.Enum):
    STATVFS_ST_RDONLY = c_sftp.SSH_FXE_STATVFS_ST_RDONLY
    STATVFS_ST_NOSUID = c_sftp.SSH_FXE_STATVFS_ST_NOSUID
class SFTP_AT(enum.Enum):
    O_RDONLY = c_sftp.O_RDONLY
    O_WRONLY = c_sftp.O_WRONLY
    O_RDWR = c_sftp.O_RDWR
    O_CREAT = c_sftp.O_CREAT
    O_EXCL = c_sftp.O_EXCL
    O_NOCTTY = c_sftp.O_NOCTTY
    O_NOFOLLOW = c_sftp.O_NOFOLLOW
    O_TRUNC = c_sftp.O_TRUNC



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from libc.stdlib cimport malloc, free
from libc.string cimport const_char

from .utils cimport to_bytes, to_str, handle_error_codes

from . cimport c_ssh


SSH_SCP_REQUEST_NEWDIR = c_ssh.ssh_scp_request_types.SSH_SCP_REQUEST_NEWDIR
SSH_SCP_REQUEST_NEWFILE = c_ssh.ssh_scp_request_types.SSH_SCP_REQUEST_NEWFILE
SSH_SCP_REQUEST_EOF = c_ssh.ssh_scp_request_types.SSH_SCP_REQUEST_EOF
SSH_SCP_REQUEST_ENDDIR = c_ssh.ssh_scp_request_types.SSH_SCP_REQUEST_ENDDIR
SSH_SCP_REQUEST_WARNING = c_ssh.ssh_scp_request_types.SSH_SCP_REQUEST_WARNING


SSH_SCP_WRITE = c_ssh.SSH_SCP_WRITE
SSH_SCP_READ = c_ssh.SSH_SCP_READ
SSH_SCP_RECURSIVE = c_ssh.SSH_SCP_RECURSIVE


cdef class SCP:

    def __cinit__(self):
        self.closed = False

    def __dealloc__(self):
        if self._scp is not NULL:
            if not self.closed:
                self.close()
            c_ssh.ssh_scp_free(self._scp)
            self._scp = NULL

    @staticmethod
    cdef SCP from_ptr(c_ssh.ssh_scp _scp, Session session):
        cdef SCP scp = SCP.__new__(SCP)
        scp._scp = _scp
        scp.session = session
        return scp

    def accept_request(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_scp_accept_request(self._scp)
        return handle_error_codes(rc, self.session._session)

    def deny_request(self, reason=None):
        cdef int rc
        cdef bytes b_reason
        cdef const char *_reason = NULL
        if reason is not None:
            b_reason = to_bytes(reason)
            _reason = b_reason
        with nogil:
            rc = c_ssh.ssh_scp_deny_request(self._scp, _reason)
        return handle_error_codes(rc, self.session._session)

    cpdef close(self):
        cdef int rc
        if self.closed:
            return 0
        with nogil:
            rc = c_ssh.ssh_scp_close(self._scp)
            if rc == 0:
                self.closed = True
        return handle_error_codes(rc, self.session._session)

    def init(self):
        cdef int rc
        with self.session._block_lock:
            with nogil:
                rc = c_ssh.ssh_scp_init(self._scp)
        return handle_error_codes(rc, self.session._session)

    def leave_directory(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_scp_leave_directory(self._scp)
        return handle_error_codes(rc, self.session._session)

    def pull_request(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_scp_pull_request(self._scp)
        return handle_error_codes(rc, self.session._session)

    def push_directory(self, dirname not None, int mode):
        cdef int rc
        cdef bytes b_dirname = to_bytes(dirname)
        cdef const char *_dirname = b_dirname
        with nogil:
            rc = c_ssh.ssh_scp_push_directory(self._scp, _dirname, mode)
        return handle_error_codes(rc, self.session._session)

    def push_file(self, filename not None, size_t size, int perms):
        cdef int rc
        cdef bytes b_filename = to_bytes(filename)
        cdef const char *_filename = b_filename
        with nogil:
            rc = c_ssh.ssh_scp_push_file(self._scp, _filename, size, perms)
        return handle_error_codes(rc, self.session._session)

    def push_file64(self, filename not None, c_ssh.uint64_t size, int perms):
        cdef int rc
        cdef bytes b_filename = to_bytes(filename)
        cdef const char *_filename = b_filename
        with nogil:
            rc = c_ssh.ssh_scp_push_file64(self._scp, _filename, size, perms)
        return handle_error_codes(rc, self.session._session)

    def read(self, c_ssh.uint32_t size=1024*1024):
        cdef int rc
        cdef bytes buf = b''
        cdef char* cbuf
        with nogil:
            cbuf = <char *>malloc(sizeof(char)*size)
            if cbuf is NULL:
                with gil:
                    raise MemoryError
            rc = c_ssh.ssh_scp_read(self._scp, cbuf, size)
        try:
            if rc > 0:
                buf = cbuf[:rc]
        finally:
            free(cbuf)
        return handle_error_codes(rc, self.session._session), buf

    def request_get_filename(self):
        cdef const char *_filename
        cdef bytes filename = b''
        with nogil:
            _filename = c_ssh.ssh_scp_request_get_filename(self._scp)
        if _filename is NULL:
            return filename
        filename = _filename
        return filename

    def request_get_permissions(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_scp_request_get_permissions(self._scp)
        return rc

    def request_get_size(self):
        cdef size_t rc
        with nogil:
            rc = c_ssh.ssh_scp_request_get_size(self._scp)
        return rc

    def request_get_size64(self):
        cdef c_ssh.uint64_t rc
        with nogil:
            rc = c_ssh.ssh_scp_request_get_size64(self._scp)
        return rc

    def request_get_warning(self):
        cdef const char *_warning
        cdef bytes warning = b''
        with nogil:
            _warning = c_ssh.ssh_scp_request_get_warning(self._scp)
        if _warning is NULL:
            return warning
        warning = _warning
        return warning

    def write(self, bytes data):
        cdef size_t size = len(data)
        cdef const_char *c_data = data
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_scp_write(self._scp, c_data, size)
        return handle_error_codes(rc, self.session._session)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

import time
import threading
import select as pyselect
from cpython cimport PyObject_AsFileDescriptor
from libc.stdlib cimport malloc, free
from libc.string cimport const_char

from .channel cimport Channel
from .connector cimport Connector
from .utils cimport to_bytes, to_str, handle_error_codes, handle_auth_error_codes, handle_sftp_error_codes
from .options cimport Option
from .key cimport SSHKey
from .sftp cimport SFTP
from .scp cimport SCP

from .exceptions import OptionError, InvalidAPIUse, ChannelOpenFailure

from .c_sftp cimport sftp_session, sftp_new, sftp_init, sftp_new_channel
from . cimport c_ssh


# SSH status flags
SSH_CLOSED = c_ssh.SSH_CLOSED
SSH_READ_PENDING = c_ssh.SSH_READ_PENDING
SSH_CLOSED_ERROR = c_ssh.SSH_CLOSED_ERROR
SSH_WRITE_PENDING = c_ssh.SSH_WRITE_PENDING


# Authentication codes
SSH_AUTH_SUCCESS = c_ssh.ssh_auth_e.SSH_AUTH_SUCCESS
SSH_AUTH_DENIED = c_ssh.ssh_auth_e.SSH_AUTH_DENIED
SSH_AUTH_PARTIAL = c_ssh.ssh_auth_e.SSH_AUTH_PARTIAL
SSH_AUTH_INFO = c_ssh.ssh_auth_e.SSH_AUTH_INFO
SSH_AUTH_AGAIN = c_ssh.ssh_auth_e.SSH_AUTH_AGAIN
SSH_AUTH_ERROR = c_ssh.ssh_auth_e.SSH_AUTH_ERROR


cdef bint _check_connected(c_ssh.ssh_session session) nogil except -1:
    if not c_ssh.ssh_is_connected(session):
        with gil:
            raise InvalidAPIUse("Session is not connected")
    return 0


cdef class Session:
    """Libssh session class providing session related functions."""

    def __cinit__(self):
        self._session = c_ssh.ssh_new()
        if self._session is NULL:
            raise MemoryError
        self._block_lock = threading.RLock()
        IF HAVE_POLL==1:
            self.c_poll_enabled = True
        ELSE:
            self.c_poll_enabled = False
        self.c_poll_use = True
        self._default_waitsockets = []
        self._waitsockets = []

    def __dealloc__(self):
        if self._session is not NULL:
            c_ssh.ssh_free(self._session)
            self._session = NULL

    cdef void _build_c_waitsocket_data(Session self) nogil:
        self._c_waitsockets[0].fd = self._sock
        self._c_waitsockets[0].events = 0
        self._c_waitsockets[0].revents = 0

    def _build_waitsocket_data(self):
        self._waitsockets = [self.sock]

    IF HAVE_POLL==1:
        cdef int poll_socket(Session self,int block_dir,int timeout) nogil:
            cdef int rc

            with nogil:
                if(block_dir & c_ssh.SSH_READ_PENDING):
                    self._c_waitsockets[0].events |= utils.POLLIN

                if(block_dir & c_ssh.SSH_WRITE_PENDING):
                    self._c_waitsockets[0].events |= utils.POLLOUT

                rc = utils.poll(self._c_waitsockets, 1, timeout)
                self._c_waitsockets[0].events = 0
            return rc

    def check_c_poll_enabled(self):
        with self._block_lock:
            return(self.c_poll_enabled==True and self.c_poll_use==True)

    def _block_call(self,_select_timeout=None):
        if _select_timeout==None:
            _select_timeout = 0.005
        with self._block_lock:
            _check_connected(self._session)
            block_direction = self.get_poll_flags()
        if block_direction==0:
            time.sleep(0.1)
            return(None)

        if self.check_c_poll_enabled()==True:
            with self._block_lock:
                return(self.poll_socket(block_direction,_select_timeout*1000))
        else:
            rfds = self._default_waitsockets
            wfds = self._default_waitsockets
            if block_direction & c_ssh.SSH_READ_PENDING:
                rfds = self._waitsockets

            if block_direction & c_ssh.SSH_WRITE_PENDING:
                wfds = self._waitsockets

            return(pyselect.select(rfds,wfds,self._default_waitsockets,_select_timeout))

    def set_socket(self, socket not None):
        """Set socket to use for session.

        Not part of libssh API but needs to be done in C to be able to
        translate python sockets to file descriptors to be used by libssh.
        """
        cdef c_ssh.socket_t _sock = PyObject_AsFileDescriptor(socket)
        cdef c_ssh.ssh_options_e fd = c_ssh.ssh_options_e.SSH_OPTIONS_FD
        cdef int rc
        self.sock = socket
        self._sock = _sock
        with nogil:
            rc = c_ssh.ssh_options_set(self._session, fd, &_sock)
        handle_error_codes(rc, self._session)
        if self.c_poll_enabled==True:
            self._build_c_waitsocket_data()
        if self.check_c_poll_enabled()==False:
            self._build_waitsocket_data()
        return rc

    def blocking_flush(self, int timeout):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_blocking_flush(self._session, timeout)
        return handle_error_codes(rc, self._session)

    def channel_new(self):
        cdef c_ssh.ssh_channel _channel
        cdef Channel channel
        with nogil:
            _check_connected(self._session)
            _channel = c_ssh.ssh_channel_new(self._session)
        if _channel is NULL:
            raise ChannelOpenFailure
        channel = Channel.from_ptr(_channel, self)
        return channel

    def sftp_new(self):
        cdef sftp_session _sftp
        cdef SFTP sftp
        with nogil:
            _check_connected(self._session)
            _sftp = sftp_new(self._session)
        if _sftp is NULL:
            return handle_sftp_error_codes(c_ssh.ssh_get_error_code(self._session), self._session)
        sftp = SFTP.from_ptr(_sftp, self)
        return sftp

    def sftp_init(self):
        """Convenience function for creating and initialising new SFTP
        session.

        Not part of libssh API. Horribly broken in nonblocking mode"""
        cdef sftp_session _sftp
        cdef SFTP sftp
        cdef int rc
        with nogil:
            _check_connected(self._session)
            _sftp = sftp_new(self._session)
        if _sftp is NULL:
            return handle_sftp_error_codes(c_ssh.ssh_get_error_code(self._session), self._session)
        with nogil:
            rc = sftp_init(_sftp)
        sftp = SFTP.from_ptr(_sftp, self)
        handle_sftp_error_codes(rc, self._session)
        return sftp

    def sftp_new_channel(self, Channel channel):
        cdef sftp_session _sftp
        cdef SFTP sftp
        cdef int rc
        with nogil:
            _check_connected(self._session)
        _sftp = sftp_new_channel(self._session,channel._channel)
        if _sftp is NULL:
            return handle_sftp_error_codes(c_ssh.ssh_get_error_code(self._session), self._session)
        with nogil:
            rc = sftp_init(_sftp)
        sftp = SFTP.from_ptr(_sftp, self)
        handle_sftp_error_codes(rc, self._session)
        return sftp

    def connect(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_connect(self._session)
        return handle_error_codes(rc, self._session)

    def disconnect(self):
        """No-op. Handled by object de-allocation."""
        # Due to bug in libssh that segfaults if session
        # is disconnected before freeing channels spawned
        # by that session - even if channels are closed.
        pass
        # if not c_ssh.ssh_is_connected(self._session):
        #     return
        # with nogil:
        #     c_ssh.ssh_disconnect(self._session)

    def connector_new(self):
        cdef c_ssh.ssh_connector _connector
        with nogil:
            _connector = c_ssh.ssh_connector_new(self._session)
        if _connector is NULL:
            return
        return Connector.from_ptr(_connector, self)

    def accept_forward(self, int timeout, int dest_port):
        cdef c_ssh.ssh_channel _channel
        with nogil:
            _check_connected(self._session)
            _channel = c_ssh.ssh_channel_accept_forward(self._session, timeout, &dest_port)
        if _channel is NULL:
            return
        return Channel.from_ptr(_channel, self)

    def cancel_forward(self, address not None, int port):
        cdef bytes b_address = to_bytes(address)
        cdef char *c_address = b_address
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_channel_cancel_forward(self._session, c_address, port)
        return handle_error_codes(rc, self._session)

    def listen_forward(self, address not None, int port, int bound_port):
        cdef bytes b_address = to_bytes(address)
        cdef char *c_address = b_address
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_channel_listen_forward(self._session, c_address, port, &bound_port)
        return handle_error_codes(rc, self._session)

    def get_disconnect_message(self):
        cdef const char *message
        cdef bytes b_message
        with nogil:
            _check_connected(self._session)
            message = c_ssh.ssh_get_disconnect_message(self._session)
        b_message = message
        return b_message

    def get_fd(self):
        cdef c_ssh.socket_t _sock
        with nogil:
            _sock = c_ssh.ssh_get_fd(self._session)
        return _sock

    def get_issue_banner(self):
        cdef char *_banner
        cdef bytes banner
        with nogil:
            _check_connected(self._session)
            _banner = c_ssh.ssh_get_issue_banner(self._session)
        if _banner is NULL:
            return
        banner = _banner
        return banner

    def get_openssh_version(self):
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_get_openssh_version(self._session)
        return rc

    def get_server_publickey(self):
        raise NotImplementedError

    def get_version(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_get_version(self._session)
        return rc

    def get_status(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_get_status(self._session)
        return rc

    def get_poll_flags(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_get_poll_flags(self._session)
        return rc

    def is_blocking(self):
        cdef bint rc
        with nogil:
            rc = c_ssh.ssh_is_blocking(self._session)
        return bool(rc)

    def is_connected(self):
        cdef bint rc
        with nogil:
            rc = c_ssh.ssh_is_connected(self._session)
        return bool(rc)

    def is_server_known(self):
        cdef bint rc
        with nogil:
            rc = c_ssh.ssh_is_server_known(self._session)
        return bool(rc)

    def copy_options(self, Session destination):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_options_copy(self._session, &destination._session)
        return handle_error_codes(rc, self._session)

    def options_getopt(self):
        raise NotImplementedError

    def options_parse_config(self, filepath):
        cdef bytes b_filepath = to_bytes(filepath)
        cdef char *c_filepath = b_filepath
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_options_parse_config(self._session, c_filepath)
        return handle_error_codes(rc, self._session)

    def options_set_port(self, int port):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_options_set(self._session, c_ssh.ssh_options_e.SSH_OPTIONS_PORT, &port)
        return handle_error_codes(rc, self._session)

    def options_set_gssapi_delegate_credentials(self, bint delegate):
        """
        Set delegating credentials to server on/off.

        :param delegate: Delegation on/off
        :type delegate: bool
        """
        with nogil:
            rc = c_ssh.ssh_options_set(self._session, c_ssh.ssh_options_e.SSH_OPTIONS_GSSAPI_DELEGATE_CREDENTIALS, &delegate)
        return handle_error_codes(rc, self._session)

    def options_set(self, Option option, value):
        """Set an option for session. This function can only be used for
        string options like host. For numeric options, port etc, use the
        individual functions.

        :param option: An SSH option object from one of
          :py:mod:`ssh.options`.
        :type option: :py:class:`ssh.options.Option`
        """
        cdef bytes b_value = to_bytes(value)
        cdef char *c_value
        cdef int rc
        c_value = b_value
        with nogil:
            rc = c_ssh.ssh_options_set(self._session, option._option, c_value)
        return handle_error_codes(rc, self._session)

    def options_get(self, Option option):
        """Get option value. This function can only be used for string optinos.
        For numeric or other options use the individual functions.
        """
        cdef char *_value
        cdef char **value = NULL
        cdef int rc
        cdef bytes b_value
        with nogil:
            rc = c_ssh.ssh_options_get(self._session, option._option, value)
        if rc < 0:
            raise OptionError
        _value = value[0]
        b_value = _value
        c_ssh.ssh_string_free_char(_value)
        return to_str(b_value)

    def options_get_port(self, unsigned int port_target):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_options_get_port(self._session, &port_target)
        return handle_error_codes(rc, self._session)

    def send_ignore(self, bytes data):
        cdef char *c_data = data
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_send_ignore(self._session, c_data)
        return handle_error_codes(rc, self._session)

    def send_debug(self, bytes message, int always_display):
        cdef char *c_message = message
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_send_debug(self._session, c_message, always_display)
        return handle_error_codes(rc, self._session)

    def gssapi_set_creds(self, creds not None):
        raise NotImplementedError

    def service_request(self, bytes service):
        cdef int rc
        cdef char *c_service = service
        with nogil:
            rc = c_ssh.ssh_service_request(self._session, c_service)
        return handle_error_codes(rc, self._session)

    # These are also excluded from Windows builds.
    IF not ON_WINDOWS:
        def set_agent_channel(self, Channel channel):
            cdef int rc
            with nogil:
                rc = c_ssh.ssh_set_agent_channel(self._session, channel._channel)
            return handle_error_codes(rc, self._session)

        def set_agent_socket(self, socket not None):
            cdef int rc
            cdef c_ssh.socket_t _sock = PyObject_AsFileDescriptor(socket)
            with nogil:
                rc = c_ssh.ssh_set_agent_socket(self._session, _sock)
            return handle_error_codes(rc, self._session)

    def set_blocking(self, int blocking):
        with nogil:
            c_ssh.ssh_set_blocking(self._session, blocking)

    def set_counters(self, scounter, rcounter):
        raise NotImplementedError

    def set_fd_except(self):
        with nogil:
            c_ssh.ssh_set_fd_except(self._session)

    def set_fd_toread(self):
        with nogil:
            c_ssh.ssh_set_fd_toread(self._session)

    def set_fd_towrite(self):
        with nogil:
            c_ssh.ssh_set_fd_towrite(self._session)

    def silent_disconnect(self):
        with nogil:
            c_ssh.ssh_silent_disconnect(self._session)

    def userauth_none(self):
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_none(self._session, NULL)
        return handle_auth_error_codes(rc, self._session)

    def userauth_list(self):
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_list(self._session, NULL)
        return handle_error_codes(rc, self._session)

    def userauth_try_publickey(self, SSHKey pubkey not None):
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_try_publickey(self._session, NULL, pubkey._key)
        return handle_auth_error_codes(rc, self._session)

    def userauth_publickey(self, SSHKey privkey not None):
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_publickey(self._session, NULL, privkey._key)
        return handle_auth_error_codes(rc, self._session)

    # ssh_userauth_agent is excluded from libssh.h on Windows.
    IF not ON_WINDOWS:
        def userauth_agent(self, username=None):
            cdef bytes b_username = to_bytes(username) if username is not None else None
            cdef char *c_username = NULL
            if b_username is not None:
                c_username = b_username
            cdef int rc
            with nogil:
                _check_connected(self._session)
                rc = c_ssh.ssh_userauth_agent(self._session, c_username)
            return handle_auth_error_codes(rc, self._session)

    def userauth_publickey_auto(self, passphrase=None):
        cdef bytes b_passphrase = to_bytes(passphrase) if passphrase is not None else None
        cdef char *c_passphrase = NULL
        if b_passphrase is not None:
            c_passphrase = b_passphrase
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_publickey_auto(self._session, NULL, c_passphrase)
        return handle_auth_error_codes(rc, self._session)

    def userauth_password(self, username=None, password=None):
        cdef bytes b_username = to_bytes(username) if username is not None else None
        cdef char *c_username = NULL
        if b_username is not None:
            c_username = b_username
        cdef bytes b_password = to_bytes(password) if password is not None else None
        cdef char *c_password = NULL
        if b_password is not None:
            c_password = b_password
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_password(self._session, c_username, c_password)
        return handle_auth_error_codes(rc, self._session)

    def userauth_kbdint(self, username=None, submethods=None):
        cdef bytes b_username = to_bytes(username) if username is not None else None
        cdef char *c_username = NULL
        if b_username is not None:
            c_username = b_username
        cdef bytes b_submethods = to_bytes(submethods) if submethods is not None else None
        cdef char *c_submethods = NULL
        if b_submethods is not None:
            c_submethods = b_submethods
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_kbdint(self._session, c_username, c_submethods)
        return handle_auth_error_codes(rc, self._session)

    def userauth_kbdint_getinstruction(self):
        cdef bytes b_instruction
        cdef const_char *_instruction
        with nogil:
            _check_connected(self._session)
            _instruction = c_ssh.ssh_userauth_kbdint_getinstruction(self._session)
        b_instruction = to_str(<char *>_instruction)
        return b_instruction

    def userauth_kbdint_getname(self):
        cdef bytes b_name
        cdef const_char *_name
        with nogil:
            _check_connected(self._session)
            _name = c_ssh.ssh_userauth_kbdint_getname(self._session)
        b_name = to_str(<char *>_name)
        return b_name

    def userauth_kbdint_getnprompts(self):
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_kbdint_getnprompts(self._session)
        return rc

    def userauth_kbdint_getprompt(self, unsigned int i, bytes echo not None):
        cdef const_char *_prompt
        cdef bytes b_prompt
        cdef char *c_echo = echo
        with nogil:
            _check_connected(self._session)
            _prompt = c_ssh.ssh_userauth_kbdint_getprompt(self._session, i, c_echo)
        b_prompt = _prompt
        return b_prompt

    def userauth_kbdint_getnanswers(self):
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_kbdint_getnanswers(self._session)
        return rc

    def userauth_kbdint_getanswer(self, unsigned int i):
        cdef const_char *_answer
        cdef bytes b_answer
        with nogil:
            _check_connected(self._session)
            _answer = c_ssh.ssh_userauth_kbdint_getanswer(self._session, i)
        b_answer = _answer
        return b_answer

    def userauth_kbdint_setanswer(self, unsigned int i, bytes answer not None):
        cdef char *c_answer = answer
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_kbdint_setanswer(self._session, i, <const_char *>(c_answer))
        return handle_auth_error_codes(rc, self._session)

    def userauth_gssapi(self):
        cdef int rc
        with nogil:
            _check_connected(self._session)
            rc = c_ssh.ssh_userauth_gssapi(self._session)
        return handle_auth_error_codes(rc, self._session)

    def write_knownhost(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_write_knownhost(self._session)
        return handle_error_codes(rc, self._session)

    def dump_knownhost(self):
        cdef const_char *_known_host
        cdef bytes b_known_host
        with nogil:
            _known_host = c_ssh.ssh_dump_knownhost(self._session)
        if _known_host is NULL:
            return
        b_known_host = _known_host
        return b_known_host

    def get_clientbanner(self):
        cdef const_char *_banner
        cdef bytes banner
        with nogil:
            _banner = c_ssh.ssh_get_clientbanner(self._session)
        if _banner is NULL:
            return
        banner = _banner
        return banner

    def get_serverbanner(self):
        cdef const_char *_banner
        cdef bytes banner
        with nogil:
            _banner = c_ssh.ssh_get_serverbanner(self._session)
        if _banner is NULL:
            return
        banner = _banner
        return banner

    def get_kex_algo(self):
        cdef const_char *_algo
        cdef bytes algo
        with nogil:
            _algo = c_ssh.ssh_get_kex_algo(self._session)
        if _algo is NULL:
            return
        algo = _algo
        return algo

    def get_cipher_in(self):
        cdef const_char *_cipher
        cdef bytes cipher
        with nogil:
            _cipher = c_ssh.ssh_get_cipher_in(self._session)
        if _cipher is NULL:
            return
        cipher = _cipher
        return cipher

    def get_cipher_out(self):
        cdef const_char *_cipher
        cdef bytes cipher
        with nogil:
            _cipher = c_ssh.ssh_get_cipher_out(self._session)
        if _cipher is NULL:
            return
        cipher = _cipher
        return cipher

    def get_hmac_in(self):
        cdef const_char *_hmac
        cdef bytes hmac
        with nogil:
            _hmac = c_ssh.ssh_get_hmac_in(self._session)
        if _hmac is NULL:
            return
        hmac = _hmac
        return hmac

    def get_hmac_out(self):
        cdef const_char *_hmac
        cdef bytes hmac
        with nogil:
            _hmac = c_ssh.ssh_get_hmac_out(self._session)
        if _hmac is NULL:
            return
        hmac = _hmac
        return hmac

    def get_error(self):
        cdef const_char *error
        cdef bytes b_error
        with nogil:
            error = c_ssh.ssh_get_error(self._session)
        if error is NULL:
            return
        b_error = error
        return to_str(b_error)

    def get_error_code(self):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_get_error_code(self._session)
        return rc

    def scp_new(self, int mode, location not None):
        """Create and initialise SCP channel"""
        cdef c_ssh.ssh_scp _scp
        cdef bytes b_location = to_bytes(location)
        cdef char *c_location = b_location
        with nogil:
            _scp = c_ssh.ssh_scp_new(self._session, mode, c_location)
            if _scp is NULL:
                with gil:
                    return handle_error_codes(c_ssh.ssh_get_error_code(self._session), self._session)
            if c_ssh.ssh_scp_init(_scp) != c_ssh.SSH_OK:
                c_ssh.ssh_scp_free(_scp)
                with gil:
                    return handle_error_codes(c_ssh.ssh_get_error_code(self._session), self._session)
        return SCP.from_ptr(_scp, self)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from .c_ssh cimport ssh_options_e


cdef class Option:
    """Class for representing an SSH option."""

    @staticmethod
    cdef Option from_option(ssh_options_e option):
        cdef Option _option = Option.__new__(Option)
        _option._option = option
        return _option

    def __eq__(self, Option other):
        return self._option == other._option

    @property
    def value(self):
        return self._option

    def __str__(self):
        return str(self._option)

    def __repr__(self):
        return self.__str__()


HOST = Option.from_option(ssh_options_e.SSH_OPTIONS_HOST)
USER = Option.from_option(ssh_options_e.SSH_OPTIONS_USER)
PORT = Option.from_option(ssh_options_e.SSH_OPTIONS_PORT)
PORT_STR = Option.from_option(ssh_options_e.SSH_OPTIONS_PORT_STR)
FD = Option.from_option(ssh_options_e.SSH_OPTIONS_FD)
SSH_DIR = Option.from_option(ssh_options_e.SSH_OPTIONS_SSH_DIR)
IDENTITY = Option.from_option(ssh_options_e.SSH_OPTIONS_IDENTITY)
ADD_IDENTITY = Option.from_option(ssh_options_e.SSH_OPTIONS_ADD_IDENTITY)
KNOWNHOSTS = Option.from_option(ssh_options_e.SSH_OPTIONS_KNOWNHOSTS)
TIMEOUT = Option.from_option(ssh_options_e.SSH_OPTIONS_TIMEOUT)
TIMEOUT_USEC = Option.from_option(ssh_options_e.SSH_OPTIONS_TIMEOUT_USEC)
SSH1 = Option.from_option(ssh_options_e.SSH_OPTIONS_SSH1)
SSH2 = Option.from_option(ssh_options_e.SSH_OPTIONS_SSH2)
LOG_VERBOSITY = Option.from_option(ssh_options_e.SSH_OPTIONS_LOG_VERBOSITY)
LOG_VERBOSITY_STR = Option.from_option(ssh_options_e.SSH_OPTIONS_LOG_VERBOSITY_STR)
CIPHERS_C_S = Option.from_option(ssh_options_e.SSH_OPTIONS_CIPHERS_C_S)
CIPHERS_S_C = Option.from_option(ssh_options_e.SSH_OPTIONS_CIPHERS_S_C)
COMPRESSION_C_S = Option.from_option(ssh_options_e.SSH_OPTIONS_COMPRESSION_C_S)
COMPRESSION_S_C = Option.from_option(ssh_options_e.SSH_OPTIONS_COMPRESSION_S_C)
PROXYCOMMAND = Option.from_option(ssh_options_e.SSH_OPTIONS_PROXYCOMMAND)
BINDADDR = Option.from_option(ssh_options_e.SSH_OPTIONS_BINDADDR)
STRICTHOSTKEYCHECK = Option.from_option(ssh_options_e.SSH_OPTIONS_STRICTHOSTKEYCHECK)
COMPRESSION = Option.from_option(ssh_options_e.SSH_OPTIONS_COMPRESSION)
COMPRESSION_LEVEL = Option.from_option(ssh_options_e.SSH_OPTIONS_COMPRESSION_LEVEL)
KEY_EXCHANGE = Option.from_option(ssh_options_e.SSH_OPTIONS_KEY_EXCHANGE)
HOSTKEYS = Option.from_option(ssh_options_e.SSH_OPTIONS_HOSTKEYS)
GSSAPI_SERVER_IDENTITY = Option.from_option(ssh_options_e.SSH_OPTIONS_GSSAPI_SERVER_IDENTITY)
GSSAPI_CLIENT_IDENTITY = Option.from_option(ssh_options_e.SSH_OPTIONS_GSSAPI_CLIENT_IDENTITY)
GSSAPI_DELEGATE_CREDENTIALS = Option.from_option(ssh_options_e.SSH_OPTIONS_GSSAPI_DELEGATE_CREDENTIALS)
HMAC_C_S = Option.from_option(ssh_options_e.SSH_OPTIONS_HMAC_C_S)
HMAC_S_C = Option.from_option(ssh_options_e.SSH_OPTIONS_HMAC_S_C)
PASSWORD_AUTH = Option.from_option(ssh_options_e.SSH_OPTIONS_PASSWORD_AUTH)
PUBKEY_AUTH = Option.from_option(ssh_options_e.SSH_OPTIONS_PUBKEY_AUTH)
KBDINT_AUTH = Option.from_option(ssh_options_e.SSH_OPTIONS_KBDINT_AUTH)
GSSAPI_AUTH = Option.from_option(ssh_options_e.SSH_OPTIONS_GSSAPI_AUTH)
GLOBAL_KNOWNHOSTS = Option.from_option(ssh_options_e.SSH_OPTIONS_GLOBAL_KNOWNHOSTS)
NODELAY = Option.from_option(ssh_options_e.SSH_OPTIONS_NODELAY)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from libc.stdlib cimport malloc, free

from .sftp cimport SFTP
from .utils cimport ssh_string_to_bytes

from .c_ssh cimport ssh_string, uint8_t, uint32_t, uint64_t
from . cimport c_sftp


cdef class SFTPAttributes:

    def __cinit__(self):
        self.self_made = False

    def __dealloc__(self):
        if self._attrs is not NULL and not self.self_made:
            c_sftp.sftp_attributes_free(self._attrs)
            self._attrs = NULL
        elif self._attrs is not NULL and self.self_made:
            free(self._attrs)
            self._attrs = NULL

    @staticmethod
    cdef SFTPAttributes from_ptr(c_sftp.sftp_attributes attrs, SFTP sftp):
        cdef SFTPAttributes _attrs = SFTPAttributes.__new__(SFTPAttributes)
        _attrs._attrs = attrs
        _attrs.sftp = sftp
        return _attrs

    @staticmethod
    def new_attrs(SFTP sftp):
        cdef SFTPAttributes attrs
        cdef c_sftp.sftp_attributes _attrs
        with nogil:
            _attrs = <c_sftp.sftp_attributes>malloc(sizeof(c_sftp.sftp_attributes_struct))
        if _attrs is NULL:
            raise MemoryError
        _attrs.name = b''
        _attrs.longname = b''
        _attrs.flags = 0
        _attrs.type = 0
        _attrs.size = 0
        _attrs.uid = 0
        _attrs.gid = 0
        _attrs.owner = b''
        _attrs.group = b''
        _attrs.permissions = 0
        _attrs.atime64 = 0
        _attrs.atime = 0
        _attrs.atime_nseconds = 0
        _attrs.createtime = 0
        _attrs.createtime_nseconds = 0
        _attrs.mtime64 = 0
        _attrs.mtime = 0
        _attrs.mtime_nseconds = 0
        _attrs.acl = NULL
        _attrs.extended_count = 0
        _attrs.extended_type = NULL
        _attrs.extended_data = NULL
        attrs = SFTPAttributes.from_ptr(_attrs, sftp)
        attrs.self_made = True
        return attrs

    @property
    def name(self):
        if self._attrs is NULL:
            return
        cdef bytes b_name = self._attrs.name
        return b_name

    @property
    def longname(self):
        if self._attrs is NULL:
            return
        cdef bytes b_longname = self._attrs.longname
        return b_longname

    @property
    def flags(self):
        return self._attrs.flags if self._attrs is not NULL else None

    @flags.setter
    def flags(self, uint32_t flags):
        self._attrs.flags = flags

    @property
    def type(self):
        return self._attrs.type if self._attrs is not NULL else None

    @type.setter
    def type(self, uint8_t _type):
        self._attrs.type = _type

    @property
    def size(self):
        return self._attrs.size if self._attrs is not NULL else None

    @size.setter
    def size(self, uint64_t size):
        self._attrs.size = size

    @property
    def uid(self):
        return self._attrs.uid if self._attrs is not NULL else None

    @uid.setter
    def uid(self, uint32_t uid):
        self._attrs.uid = uid

    @property
    def gid(self):
        return self._attrs.gid if self._attrs is not NULL else None

    @gid.setter
    def gid(self, uint32_t gid):
        self._attrs.gid = gid

    @property
    def owner(self):
        if self._attrs is NULL:
            return
        cdef bytes b_owner = self._attrs.owner if self._attrs.owner is not NULL else None
        return b_owner

    @property
    def group(self):
        if self._attrs is NULL:
            return
        cdef bytes b_group = self._attrs.group if self._attrs.group is not NULL else None
        return b_group

    @property
    def permissions(self):
        return self._attrs.permissions if self._attrs is not NULL else None

    @permissions.setter
    def permissions(self, uint32_t permissions):
        self._attrs.permissions = permissions

    @property
    def atime64(self):
        return self._attrs.atime64 if self._attrs is not NULL else None

    @atime64.setter
    def atime64(self, uint64_t atime):
        self._attrs.atime64 = atime

    @property
    def atime(self):
        return self._attrs.atime if self._attrs is not NULL else None

    @atime.setter
    def atime(self, uint32_t atime):
        self._attrs.atime = atime

    @property
    def atime_nseconds(self):
        return self._attrs.atime_nseconds if self._attrs is not NULL else None

    @atime_nseconds.setter
    def atime_nseconds(self, uint32_t nseconds):
        self._attrs.atime_nseconds = nseconds

    @property
    def createtime(self):
        return self._attrs.createtime if self._attrs is not NULL else None

    @createtime.setter
    def createtime(self, uint64_t createtime):
        self._attrs.createtime = createtime

    @property
    def createtime_nseconds(self):
        return self._attrs.createtime_nseconds if self._attrs is not NULL else None

    @createtime_nseconds.setter
    def createtime_nseconds(self, uint32_t nseconds):
        self._attrs.createtime_nseconds = nseconds

    @property
    def mtime64(self):
        return self._attrs.mtime64 if self._attrs is not NULL else None

    @mtime64.setter
    def mtime64(self, uint64_t mtime):
        self._attrs.mtime64 = mtime

    @property
    def mtime(self):
        return self._attrs.mtime if self._attrs is not NULL else None

    @mtime.setter
    def mtime(self, uint32_t mtime):
        self._attrs.mtime = mtime

    @property
    def mtime_nseconds(self):
        return self._attrs.mtime_nseconds if self._attrs is not NULL else None

    @mtime_nseconds.setter
    def mtime_nseconds(self, uint32_t nseconds):
        self._attrs.mtime_nseconds = nseconds

    @property
    def acl(self):
        if self._attrs is NULL:
            return
        return ssh_string_to_bytes(self._attrs.acl)

    @property
    def extended_count(self):
        return self._attrs.extended_count if self._attrs is not NULL else None

    @extended_count.setter
    def extended_count(self, uint32_t count):
        self._attrs.extended_count = count

    @property
    def extended_type(self):
        if self._attrs is NULL:
            return
        return ssh_string_to_bytes(self._attrs.extended_type)

    @property
    def extended_data(self):
        if self._attrs is NULL:
            return
        return ssh_string_to_bytes(self._attrs.extended_data)



# This file is part of RedLibSSH.
# Copyright (C) 2018-2020 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from libc.string cimport const_char

from .utils cimport to_str, to_bytes
from .c_ssh cimport ssh_keytypes_e, ssh_key_type_to_char, ssh_key_type_from_name


cdef class KeyType:
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_UNKNOWN

    @property
    def value(self):
        return self._type

    def __str__(self):
        cdef const_char *c_type
        c_type = ssh_key_type_to_char(self._type)
        if c_type is not NULL:
            return to_str(c_type)
        return "unknown"

    def __repr__(self):
        return self.__str__()


cdef class UnknownKey(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_UNKNOWN


cdef class DSSKey(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_DSS


cdef class RSAKey(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_RSA


cdef class RSA1Key(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_RSA1


cdef class ECDSAKey(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_ECDSA


cdef class DSSCert01Key(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_DSS_CERT01


cdef class RSACert01Key(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_RSA_CERT01


cdef class ECDSA_P256(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P256


cdef class ECDSA_P384(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P384


cdef class ECDSA_P521(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P521


cdef class ECDSA_P256_CERT01(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P256_CERT01


cdef class ECDSA_P384_CERT01(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P384_CERT01


cdef class ECDSA_P521_CERT01(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P521_CERT01


cdef class ED25519Key(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_ED25519


cdef class ED25519_CERT01(KeyType):
    def __cinit__(self):
        self._type = ssh_keytypes_e.SSH_KEYTYPE_ED25519_CERT01


cdef KeyType from_keytype(ssh_keytypes_e _type):
    if _type == ssh_keytypes_e.SSH_KEYTYPE_UNKNOWN:
        return UnknownKey()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_DSS:
        return DSSKey()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_RSA:
        return RSAKey()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_RSA1:
        return RSA1Key()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_ECDSA:
        return ECDSAKey()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_DSS_CERT01:
        return DSSCert01Key()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_RSA_CERT01:
        return RSACert01Key()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P256:
        return ECDSA_P256()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P384:
        return ECDSA_P384()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P521:
        return ECDSA_P521()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P256_CERT01:
        return ECDSA_P256_CERT01()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P384_CERT01:
        return ECDSA_P384_CERT01()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_ECDSA_P521_CERT01:
        return ECDSA_P521_CERT01()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_ED25519:
        return ED25519Key()
    elif _type == ssh_keytypes_e.SSH_KEYTYPE_ED25519_CERT01:
        return ED25519_CERT01()
    else:
        raise ValueError("Unknown keytype %s", _type)


def key_type_from_name(key_name):
    cdef ssh_keytypes_e _type
    cdef bytes b_key_name = to_bytes(key_name)
    cdef const char *_key_name = b_key_name
    with nogil:
        _type = ssh_key_type_from_name(_key_name)
    return from_keytype(_type)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from cpython cimport PyObject_AsFileDescriptor

from .utils cimport handle_error_codes

from .connector cimport Connector
from .session cimport Session

from . cimport c_ssh


cdef class Event:

    def __cinit__(self):
        self._event = c_ssh.ssh_event_new()
        if self._event is NULL:
            raise MemoryError

    def __dealloc__(self):
        if self._event is not NULL:
            c_ssh.ssh_event_free(self._event)
            self._event = NULL

    @property
    def socket(self):
        return self._sock

    @staticmethod
    cdef Event from_ptr(c_ssh.ssh_event _event):
        cdef Event event = Event.__new__(Event)
        event._event = _event
        return event

    @staticmethod
    cdef int event_callback(c_ssh.socket_t fd, int revent, void *userdata):
        try:
            func = <object>userdata
            return func()
        except Exception:
            # TODO - pass back exception
            return -1

    def add_fd(self, sock, short events, callback=None):
        cdef c_ssh.socket_t _sock = PyObject_AsFileDescriptor(sock)
        cdef c_ssh.ssh_event_callback cb = <c_ssh.ssh_event_callback>&Event.event_callback
        cdef int rc
        cdef void *_cb = NULL if callback is None else <void *>callback
        rc = c_ssh.ssh_event_add_fd(self._event, _sock, events, cb, _cb)
        if rc == 0:
            self._sock = sock
        return rc

    def remove_fd(self, socket):
        cdef c_ssh.socket_t _sock = PyObject_AsFileDescriptor(socket)
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_event_remove_fd(self._event, _sock)
        if rc == 0:
            self._sock = None
        return rc

    def add_session(self, Session session):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_event_add_session(self._event, session._session)
        handle_error_codes(rc, session._session)
        self.session = session
        return rc

    def add_connector(self, Connector connector):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_event_add_connector(self._event, connector._connector)
        if rc == 0:
            self.connector = connector
        return handle_error_codes(rc, connector.session._session)

    def dopoll(self, int timeout):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_event_dopoll(self._event, timeout)
        return rc

    def remove_session(self, Session session):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_event_remove_session(self._event, session._session)
        handle_error_codes(rc, session._session)
        self.session = None
        return rc

    def remove_connector(self, Connector connector):
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_event_remove_connector(self._event, connector._connector)
        if rc == 0:
            self.connector = None
        return handle_error_codes(rc, connector.session._session)



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from .c_ssh cimport ssh_session, ssh_channel, ssh_buffer, ssh_string, uint32_t, uint8_t, ssh_key_struct, ssh_auth_callback, ssh_message

cdef extern from "libssh/callbacks.h" nogil:
    ctypedef void(*ssh_callback_int)(int code, void *user)
    ctypedef int(*ssh_callback_data)(const void *data, size_t len, void *user)
    ctypedef void(*ssh_callback_int_int)(int code, int errno_code, void *user)
    ctypedef int(*ssh_message_callback)(ssh_session, ssh_message message, void *user)
    ctypedef int(*ssh_channel_callback_int)(ssh_channel channel, int code, void *user)
    ctypedef int(*ssh_channel_callback_data)(ssh_channel channel, int code, void *data, size_t len, void *user)

    ctypedef void(*ssh_log_callback)(ssh_session session, int priority,   const char *message, void *userdata)
    ctypedef void(*ssh_logging_callback)(int priority, const char *function, const char *buffer, void *userdata)
    ctypedef void(*ssh_status_callback)(ssh_session session, float status, void *userdata)
    ctypedef void(*ssh_global_request_callback)(ssh_session session, ssh_message message, void *userdata)
    ctypedef ssh_channel(*ssh_channel_open_request_x11_callback)(ssh_session session, const char * originator_address,int originator_port, void *userdata)
    ctypedef ssh_channel(*ssh_channel_open_request_auth_agent_callback)(ssh_session session, void *userdata)
    struct ssh_callbacks_struct:
        size_t size
        void *userdata
        ssh_auth_callback auth_function
        ssh_log_callback log_function
        void(*connect_status_function)(void *userdata, float status)
        ssh_global_request_callback global_request_function
        ssh_channel_open_request_x11_callback channel_open_request_x11_function
        ssh_channel_open_request_auth_agent_callback channel_open_request_auth_agent_function
    ctypedef ssh_callbacks_struct *ssh_callbacks

    ctypedef int(*ssh_auth_password_callback)(ssh_session session, const char *user, const char *password,void *userdata)
    ctypedef int(*ssh_auth_none_callback)(ssh_session session, const char *user, void *userdata)
    ctypedef int(*ssh_auth_gssapi_mic_callback)(ssh_session session, const char *user, const char *principal,void *userdata)
    ctypedef int(*ssh_auth_pubkey_callback)(ssh_session session, const char *user, ssh_key_struct *pubkey,char signature_state, void *userdata)
    ctypedef int(*ssh_service_request_callback)(ssh_session session, const char *service, void *userdata)
    ctypedef ssh_channel(*ssh_channel_open_request_session_callback)(ssh_session session, void *userdata)

    ctypedef ssh_string(*void)(ssh_session, const char*, int, ssh_string *, void *)
    ctypedef int(*ssh_gssapi_accept_sec_ctx_callback)(ssh_session session, ssh_string input_token,ssh_string *output_token, void *userdata)
    ctypedef int(*ssh_gssapi_verify_mic_callback)(ssh_session session, ssh_string mic, void *mic_buffer,size_t mic_buffer_size, void *userdata)

    ctypedef ssh_string(*ssh_gssapi_select_oid_callback)(ssh_session session, const char *user,int n_oid, ssh_string *oids, void *userdata)

    struct ssh_server_callbacks_struct:
        size_t size
        void *userdata
        ssh_auth_password_callback auth_password_function
        ssh_auth_none_callback auth_none_function
        ssh_auth_gssapi_mic_callback auth_gssapi_mic_function
        ssh_auth_pubkey_callback auth_pubkey_function
        ssh_service_request_callback service_request_function
        ssh_channel_open_request_session_callback channel_open_request_session_function
        ssh_gssapi_select_oid_callback gssapi_select_oid_function
        ssh_gssapi_accept_sec_ctx_callback gssapi_accept_sec_ctx_function
        ssh_gssapi_verify_mic_callback gssapi_verify_mic_function
    ctypedef ssh_server_callbacks_struct *ssh_server_callbacks
    int ssh_set_server_callbacks(ssh_session session, ssh_server_callbacks cb)
    void ssh_callbacks_init(void *cb)
    struct ssh_socket_callbacks_struct:
        void *userdata
        ssh_callback_data data
        ssh_callback_int controlflow
        ssh_callback_int_int exception
        ssh_callback_int_int connected
    ctypedef ssh_socket_callbacks_struct *ssh_socket_callbacks

    enum:
        SSH_SOCKET_FLOW_WRITEWILLBLOCK
        SSH_SOCKET_FLOW_WRITEWONTBLOCK
        SSH_SOCKET_EXCEPTION_EOF
        SSH_SOCKET_EXCEPTION_ERROR
        SSH_SOCKET_CONNECTED_OK
        SSH_SOCKET_CONNECTED_ERROR
        SSH_SOCKET_CONNECTED_TIMEOUT
    ctypedef int(*ssh_packet_callback)(ssh_session session, uint8_t type, ssh_buffer packet, void *user)
    enum:
        SSH_PACKET_USED
        SSH_PACKET_NOT_USED

    struct ssh_packet_callbacks_struct:
        uint8_t start
        uint8_t n_callbacks
        ssh_packet_callback *callbacks
        void *user
    ctypedef ssh_packet_callbacks_struct *ssh_packet_callbacks
    int ssh_set_callbacks(ssh_session session, ssh_callbacks cb)
    ctypedef int(*ssh_channel_data_callback)(ssh_session session, ssh_channel channel, void *data, uint32_t len,int is_stderr, void *userdata)

    ctypedef void(*ssh_channel_eof_callback)(ssh_session session, ssh_channel channel, void *userdata)
    ctypedef void(*ssh_channel_close_callback)(ssh_session session, ssh_channel channel, void *userdata)
    ctypedef void(*ssh_channel_signal_callback)(ssh_session session, ssh_channel channel,const char *signal, void *userdata)
    ctypedef void(*ssh_channel_exit_status_callback)(ssh_session session, ssh_channel channel,int exit_status, void *userdata)
    ctypedef void(*ssh_channel_exit_signal_callback)(ssh_session session, ssh_channel channel, const char *signal, int core,const char *errmsg, const char *lang, void *userdata)
    ctypedef int(*ssh_channel_pty_request_callback)(ssh_session session, ssh_channel channel, const char *term,int width, int height, int pxwidth, int pwheight, void *userdata)
    ctypedef int(*ssh_channel_shell_request_callback)(ssh_session session, ssh_channel channel, void *userdata)
    ctypedef void(*ssh_channel_auth_agent_req_callback)(ssh_session session, ssh_channel channel, void *userdata)
    ctypedef void(*ssh_channel_x11_req_callback)(ssh_session session, ssh_channel channel,int single_connection, const char *auth_protocol,const char *auth_cookie, uint32_t screen_number, void *userdata)
    ctypedef int(*ssh_channel_pty_window_change_callback)(ssh_session session, ssh_channel channel, int width, int height,int pxwidth, int pwheight, void *userdata)
    ctypedef int(*ssh_channel_exec_request_callback)(ssh_session session, ssh_channel channel, const char *command,void *userdata)
    ctypedef int(*ssh_channel_env_request_callback)(ssh_session session, ssh_channel channel, const char *env_name,const char *env_value, void *userdata)
    ctypedef int(*ssh_channel_subsystem_request_callback)(ssh_session session, ssh_channel channel, const char *subsystem,void *userdata)
    ctypedef int(*ssh_channel_write_wontblock_callback)(ssh_session session, ssh_channel channel,size_t bytes, void *userdata)
    struct ssh_channel_callbacks_struct:
        size_t size
        void *userdata
        ssh_channel_data_callback channel_data_function
        ssh_channel_eof_callback channel_eof_function
        ssh_channel_close_callback channel_close_function
        ssh_channel_signal_callback channel_signal_function
        ssh_channel_exit_status_callback channel_exit_status_function
        ssh_channel_exit_signal_callback channel_exit_signal_function
        ssh_channel_pty_request_callback channel_pty_request_function
        ssh_channel_shell_request_callback channel_shell_request_function
        ssh_channel_auth_agent_req_callback channel_auth_agent_req_function
        ssh_channel_x11_req_callback channel_x11_req_function
        ssh_channel_pty_window_change_callback \
            channel_pty_window_change_function
        ssh_channel_exec_request_callback channel_exec_request_function
        ssh_channel_env_request_callback channel_env_request_function
        ssh_channel_subsystem_request_callback \
            channel_subsystem_request_function
        ssh_channel_write_wontblock_callback channel_write_wontblock_function

    ctypedef ssh_channel_callbacks_struct *ssh_channel_callbacks
    int ssh_set_channel_callbacks(ssh_channel channel, ssh_channel_callbacks cb)
    int ssh_add_channel_callbacks(ssh_channel channel, ssh_channel_callbacks cb)
    int ssh_remove_channel_callbacks(ssh_channel channel, ssh_channel_callbacks cb)
    ctypedef int(*ssh_thread_callback)(void **lock)
    ctypedef unsigned long(*ssh_thread_id_callback)()
    struct ssh_threads_callbacks_struct:
        const char *type
        ssh_thread_callback mutex_init
        ssh_thread_callback mutex_destroy
        ssh_thread_callback mutex_lock
        ssh_thread_callback mutex_unlock
        ssh_thread_id_callback thread_id
    int ssh_threads_set_callbacks(ssh_threads_callbacks_struct *cb)
    ssh_threads_callbacks_struct *ssh_threads_get_pthread()
    ssh_threads_callbacks_struct *ssh_threads_get_noop()
    int ssh_set_log_callback(ssh_logging_callback cb)
    ssh_logging_callback ssh_get_log_callback()



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

cdef extern from "libssh/ssh2.h" nogil:
    enum:
        SSH2_MSG_DISCONNECT
        SSH2_MSG_IGNORE
        SSH2_MSG_UNIMPLEMENTED
        SSH2_MSG_DEBUG
        SSH2_MSG_SERVICE_REQUEST
        SSH2_MSG_SERVICE_ACCEPT
        SSH2_MSG_KEXINIT
        SSH2_MSG_NEWKEYS
        SSH2_MSG_KEXDH_INIT
        SSH2_MSG_KEXDH_REPLY
        SSH2_MSG_KEX_ECDH_INIT
        SSH2_MSG_KEX_ECDH_REPLY
        SSH2_MSG_ECMQV_INIT
        SSH2_MSG_ECMQV_REPLY
        SSH2_MSG_KEX_DH_GEX_REQUEST_OLD
        SSH2_MSG_KEX_DH_GEX_GROUP
        SSH2_MSG_KEX_DH_GEX_INIT
        SSH2_MSG_KEX_DH_GEX_REPLY
        SSH2_MSG_KEX_DH_GEX_REQUEST
        SSH2_MSG_USERAUTH_REQUEST
        SSH2_MSG_USERAUTH_FAILURE
        SSH2_MSG_USERAUTH_SUCCESS
        SSH2_MSG_USERAUTH_BANNER
        SSH2_MSG_USERAUTH_PK_OK
        SSH2_MSG_USERAUTH_PASSWD_CHANGEREQ
        SSH2_MSG_USERAUTH_INFO_REQUEST
        SSH2_MSG_USERAUTH_GSSAPI_RESPONSE
        SSH2_MSG_USERAUTH_INFO_RESPONSE
        SSH2_MSG_USERAUTH_GSSAPI_TOKEN
        SSH2_MSG_USERAUTH_GSSAPI_EXCHANGE_COMPLETE
        SSH2_MSG_USERAUTH_GSSAPI_ERROR
        SSH2_MSG_USERAUTH_GSSAPI_ERRTOK
        SSH2_MSG_USERAUTH_GSSAPI_MIC
        SSH2_MSG_GLOBAL_REQUEST
        SSH2_MSG_REQUEST_SUCCESS
        SSH2_MSG_REQUEST_FAILURE
        SSH2_MSG_CHANNEL_OPEN
        SSH2_MSG_CHANNEL_OPEN_CONFIRMATION
        SSH2_MSG_CHANNEL_OPEN_FAILURE
        SSH2_MSG_CHANNEL_WINDOW_ADJUST
        SSH2_MSG_CHANNEL_DATA
        SSH2_MSG_CHANNEL_EXTENDED_DATA
        SSH2_MSG_CHANNEL_EOF
        SSH2_MSG_CHANNEL_CLOSE
        SSH2_MSG_CHANNEL_REQUEST
        SSH2_MSG_CHANNEL_SUCCESS
        SSH2_MSG_CHANNEL_FAILURE
        SSH2_DISCONNECT_HOST_NOT_ALLOWED_TO_CONNECT
        SSH2_DISCONNECT_PROTOCOL_ERROR
        SSH2_DISCONNECT_KEY_EXCHANGE_FAILED
        SSH2_DISCONNECT_HOST_AUTHENTICATION_FAILED
        SSH2_DISCONNECT_RESERVED
        SSH2_DISCONNECT_MAC_ERROR
        SSH2_DISCONNECT_COMPRESSION_ERROR
        SSH2_DISCONNECT_SERVICE_NOT_AVAILABLE
        SSH2_DISCONNECT_PROTOCOL_VERSION_NOT_SUPPORTED
        SSH2_DISCONNECT_HOST_KEY_NOT_VERIFIABLE
        SSH2_DISCONNECT_CONNECTION_LOST
        SSH2_DISCONNECT_BY_APPLICATION
        SSH2_DISCONNECT_TOO_MANY_CONNECTIONS
        SSH2_DISCONNECT_AUTH_CANCELLED_BY_USER
        SSH2_DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE
        SSH2_DISCONNECT_ILLEGAL_USER_NAME
        SSH2_OPEN_ADMINISTRATIVELY_PROHIBITED
        SSH2_OPEN_CONNECT_FAILED
        SSH2_OPEN_UNKNOWN_CHANNEL_TYPE
        SSH2_OPEN_RESOURCE_SHORTAGE
        SSH2_EXTENDED_DATA_STDERR



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-130

from libc.string cimport const_char

from .keytypes cimport from_keytype, KeyType
from .utils cimport to_str, to_bytes

from .exceptions import KeyExportError, KeyImportError, KeyGenerationError

from . cimport c_ssh


cdef class SSHKey:

    @staticmethod
    cdef SSHKey from_ptr(c_ssh.ssh_key key):
        cdef SSHKey _key = SSHKey.__new__(SSHKey)
        _key._key = key
        return _key

    def __cinit__(self):
        self._key = c_ssh.ssh_key_new()
        if self._key is NULL:
            raise MemoryError

    def __dealloc__(self):
        if self._key is not NULL:
            c_ssh.ssh_key_free(self._key)
            self._key = NULL

    def is_private(self):
        cdef bint rc
        with nogil:
            rc = c_ssh.ssh_key_is_private(self._key)
        return bool(rc)

    def is_public(self):
        cdef bint rc
        with nogil:
            rc = c_ssh.ssh_key_is_public(self._key)
        return bool(rc)

    def __eq__(self, SSHKey other):
        cdef bint is_private
        cdef bint equal
        with nogil:
            is_private = c_ssh.ssh_key_is_private(self._key)
            equal = c_ssh.ssh_key_cmp(self._key, other._key, c_ssh.ssh_keycmp_e.SSH_KEY_CMP_PRIVATE) if is_private else c_ssh.ssh_key_cmp(self._key, other._key, c_ssh.ssh_keycmp_e.SSH_KEY_CMP_PUBLIC)
        return bool(not equal)

    def key_type(self):
        cdef c_ssh.ssh_keytypes_e _type
        if self._key is NULL:
            return
        _type = c_ssh.ssh_key_type(self._key)
        return from_keytype(_type)

    def ecdsa_name(self):
        cdef const_char *c_name
        cdef bytes b_name
        with nogil:
            c_name = c_ssh.ssh_pki_key_ecdsa_name(self._key)
        b_name = c_name
        return to_str(b_name)

    def export_privkey_file(self, filepath, passphrase=None):
        cdef bytes b_passphrase
        cdef bytes b_filepath = to_bytes(filepath)
        cdef const_char *c_passphrase = NULL
        cdef const_char *c_filepath = b_filepath
        cdef int rc
        if passphrase is not None:
            b_passphrase = to_bytes(passphrase)
            c_passphrase = b_passphrase
        with nogil:
            rc = c_ssh.ssh_pki_export_privkey_file(self._key, c_passphrase, NULL, NULL, c_filepath)
        if rc != c_ssh.SSH_OK:
            raise KeyExportError

    def export_privkey_to_pubkey(self):
        cdef SSHKey pub_key
        cdef c_ssh.ssh_key _pub_key
        cdef int rc
        with nogil:
            rc = c_ssh.ssh_pki_export_privkey_to_pubkey(self._key, &_pub_key)
        if rc != c_ssh.SSH_OK:
            raise KeyExportError
        pub_key = SSHKey.from_ptr(_pub_key)
        return pub_key

    def export_pubkey_base64(self):
        cdef char *_key
        cdef int rc
        cdef bytes b_key
        cdef size_t key_len
        with nogil:
            rc = c_ssh.ssh_pki_export_pubkey_base64(self._key, &_key)
            if rc != c_ssh.SSH_OK:
                with gil:
                    raise KeyExportError
        b_key = _key
        c_ssh.ssh_string_free_char(_key)
        return b_key


def generate(KeyType key_type, int bits):
    cdef SSHKey key
    cdef c_ssh.ssh_key _key
    cdef int rc
    with nogil:
        rc = c_ssh.ssh_pki_generate(key_type._type, bits, &_key)
    if rc != c_ssh.SSH_OK:
        raise KeyGenerationError
    key = SSHKey.from_ptr(_key)
    return key


def import_privkey_base64(bytes b64_key, passphrase=b''):
    cdef const_char *c_key = b64_key
    cdef bytes b_passphrase
    cdef const_char *c_passphrase = NULL
    cdef int rc
    cdef SSHKey key
    cdef c_ssh.ssh_key _key
    if passphrase is not None:
        b_passphrase = to_bytes(passphrase)
        c_passphrase = b_passphrase
    with nogil:
        rc = c_ssh.ssh_pki_import_privkey_base64(c_key, c_passphrase, NULL, NULL, &_key)
    if rc != c_ssh.SSH_OK:
        raise KeyImportError
    key = SSHKey.from_ptr(_key)
    return key


def import_privkey_file(filepath, passphrase=b''):
    cdef bytes b_passphrase
    cdef bytes b_filepath = to_bytes(filepath)
    cdef const_char *c_passphrase = NULL
    cdef const_char *c_filepath = b_filepath
    cdef int rc
    cdef SSHKey key
    cdef c_ssh.ssh_key _key
    if passphrase is not None:
        b_passphrase = to_bytes(passphrase)
        c_passphrase = b_passphrase
    with nogil:
        rc = c_ssh.ssh_pki_import_privkey_file(c_filepath, c_passphrase, NULL, NULL, &_key)
    if rc != c_ssh.SSH_OK:
        raise KeyImportError
    key = SSHKey.from_ptr(_key)
    return key


def import_pubkey_base64(bytes b64_key, KeyType key_type):
    cdef const_char *c_key = b64_key
    cdef int rc
    cdef SSHKey key
    cdef c_ssh.ssh_key _key
    with nogil:
        rc = c_ssh.ssh_pki_import_pubkey_base64(c_key, key_type._type, &_key)
    if rc != c_ssh.SSH_OK:
        raise KeyImportError
    key = SSHKey.from_ptr(_key)
    return key


def import_pubkey_file(filepath):
    cdef bytes b_filepath = to_bytes(filepath)
    cdef const_char *c_filepath = b_filepath
    cdef int rc
    cdef SSHKey key
    cdef c_ssh.ssh_key _key
    with nogil:
        rc = c_ssh.ssh_pki_import_pubkey_file(c_filepath, &_key)
    if rc != c_ssh.SSH_OK:
        raise KeyImportError
    key = SSHKey.from_ptr(_key)
    return key


def import_cert_base64(bytes b64_cert, KeyType key_type):
    cdef const_char *c_key = b64_cert
    cdef int rc
    cdef SSHKey key
    cdef c_ssh.ssh_key _key
    with nogil:
        rc = c_ssh.ssh_pki_import_cert_base64(c_key, key_type._type, &_key)
    if rc != c_ssh.SSH_OK:
        raise KeyImportError
    key = SSHKey.from_ptr(_key)
    return key


def import_cert_file(filepath):
    cdef bytes b_filepath = to_bytes(filepath)
    cdef const_char *c_filepath = b_filepath
    cdef int rc
    cdef SSHKey key
    cdef c_ssh.ssh_key _key
    with nogil:
        rc = c_ssh.ssh_pki_import_cert_file(c_filepath, &_key)
    if rc != c_ssh.SSH_OK:
        raise KeyImportError
    key = SSHKey.from_ptr(_key)
    return key


def copy_cert_to_privkey(SSHKey cert_key, SSHKey priv_key):
    if priv_key.is_private() is False:
        raise KeyImportError
    cdef c_ssh.ssh_key _priv_key = priv_key._key
    cdef c_ssh.ssh_key _cert_key = cert_key._key
    with nogil:
        rc = c_ssh.ssh_pki_copy_cert_to_privkey(_cert_key, _priv_key)
    if rc != c_ssh.SSH_OK:
        raise KeyImportError



# This file is part of RedLibSSH.
# Copyright (C) 2018 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

from select import select

from cpython.version cimport PY_MAJOR_VERSION

from . cimport c_ssh
from .c_ssh cimport ssh_error_types_e, ssh_get_error, ssh_auth_e, SSH_OK, SSH_ERROR, SSH_AGAIN, SSH_EOF, ssh_session, ssh_string, ssh_string_get_char, ssh_string_free, ssh_string_len, SSH_READ_PENDING, SSH_WRITE_PENDING
from .c_sftp cimport SSH_FX_OK, SSH_FX_EOF, SSH_FX_NO_SUCH_FILE, SSH_FX_PERMISSION_DENIED, SSH_FX_FAILURE, SSH_FX_BAD_MESSAGE, SSH_FX_NO_CONNECTION, SSH_FX_CONNECTION_LOST, SSH_FX_OP_UNSUPPORTED, SSH_FX_INVALID_HANDLE, SSH_FX_NO_SUCH_PATH, SSH_FX_FILE_ALREADY_EXISTS, SSH_FX_WRITE_PROTECT, SSH_FX_NO_MEDIA

from .exceptions import OtherError, AuthenticationPartial, AuthenticationDenied, AuthenticationError, SSHError, EOF, SFTPError

IF HAVE_POLL==1:
    from .utils cimport POLLIN, POLLOUT
    pollin=POLLIN
    pollout=POLLOUT

ENCODING='utf-8'

cdef bytes to_bytes(_str):
    if isinstance(_str, bytes):
        return _str
    elif isinstance(_str, unicode):
        return _str.encode(ENCODING)
    return _str


cdef object to_str(const char *c_str):
    _len = len(c_str)
    if PY_MAJOR_VERSION < 3:
        return c_str[:_len]
    return c_str[:_len].decode(ENCODING)


cdef object to_str_len(const char *c_str, int length):
    if PY_MAJOR_VERSION < 3:
        return c_str[:length]
    return c_str[:length].decode(ENCODING)


cdef bytes ssh_string_to_bytes(ssh_string _str):
    if _str is NULL:
        return
    cdef const char *c_str
    cdef size_t str_len
    cdef bytes b_str = None
    with nogil:
        str_len = ssh_string_len(_str)
        c_str = ssh_string_get_char(_str)
    if c_str is NULL:
        raise MemoryError
    try:
        b_str = c_str[:str_len]
        return b_str
    finally:
        ssh_string_free(_str)

def set_log_level(level):
    cdef int c_level = int(level)
    cdef int rc
    rc = c_ssh.ssh_set_log_level(c_level)
    return(rc)

def get_log_level():
    cdef int rc
    rc = c_ssh.ssh_get_log_level()
    return(rc)

def wait_socket(session not None, sock not None, timeout=None):
    """Helper function for testing non-blocking mode.

    This function blocks the calling thread for <timeout> seconds -
    to be used only for testing purposes.
    """
    cdef int directions = session.get_poll_flags()
    if directions == 0:
        return 0
    readfds = (sock,) if (directions & SSH_READ_PENDING) else ()
    writefds = (sock,) if (directions & SSH_WRITE_PENDING) else ()
    select(readfds, writefds, (), timeout)


cdef int handle_error_codes(int errcode, ssh_session session) except -1:
    if errcode == SSH_OK:
        return SSH_OK
    elif errcode == SSH_ERROR:
        raise SSHError(errcode, ssh_get_error(session))
    elif errcode == SSH_EOF:
        raise EOF(ssh_get_error(session))
    elif errcode == SSH_AGAIN:
        return SSH_AGAIN
    else:
        if errcode < 0:
            raise OtherError(ssh_get_error(session))
        return errcode


cdef int handle_auth_error_codes(int errcode, ssh_session session) except -1:
    if errcode == ssh_auth_e.SSH_AUTH_SUCCESS:
        return ssh_auth_e.SSH_AUTH_SUCCESS
    elif errcode == ssh_auth_e.SSH_AUTH_DENIED:
        raise AuthenticationDenied(ssh_get_error(session))
    elif errcode == ssh_auth_e.SSH_AUTH_ERROR:
        raise AuthenticationError(ssh_get_error(session))
    elif errcode == ssh_auth_e.SSH_AUTH_PARTIAL:
        raise AuthenticationPartial(ssh_get_error(session))
    elif errcode == ssh_auth_e.SSH_AUTH_AGAIN:
        return ssh_auth_e.SSH_AUTH_AGAIN
    else:
        return handle_error_codes(errcode, session)


cdef int handle_sftp_error_codes(int errcode, ssh_session session) except -1:
    if errcode == SSH_FX_OK:
        return SSH_FX_OK
    elif errcode == SSH_FX_EOF:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_NO_SUCH_FILE:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_PERMISSION_DENIED:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_FAILURE:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_BAD_MESSAGE:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_NO_CONNECTION:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_CONNECTION_LOST:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_OP_UNSUPPORTED:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_INVALID_HANDLE:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_NO_SUCH_PATH:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_FILE_ALREADY_EXISTS:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_WRITE_PROTECT:
        raise SFTPError(ssh_get_error(session))
    elif errcode == SSH_FX_NO_MEDIA:
        raise SFTPError(ssh_get_error(session))
    else:
        return handle_error_codes(errcode, session)



# This file is part of RedLibSSH.
#
# Copyright (C) 2017-2020 Panos Kittenis
# Copyright (C) 2022 Red-M
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, version 2.1.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

from .c_ssh cimport ssh_session, ssh_string

cdef bytes to_bytes(_str)
cdef object to_str(const char *c_str)
cdef object to_str_len(const char *c_str, int length)
cdef int handle_error_codes(int errcode, ssh_session) except -1
cdef int handle_auth_error_codes(int errcode, ssh_session) except -1
cdef int handle_sftp_error_codes(int errcode, ssh_session) except -1
cdef bytes ssh_string_to_bytes(ssh_string _str)

IF HAVE_POLL==1:
    cdef extern from "poll.h" nogil:
        ctypedef unsigned int nfds_t
        enum:
            POLLIN
            POLLPRI
            POLLOUT
            POLLERR
            POLLHUP
            POLLNVAL
        struct pollfd:
            int fd
            short events
            short revents
        int poll(pollfd *fds, nfds_t nfds, int timeout)



import os

from sys import stderr
from subprocess import check_call
from glob import glob
from shutil import copy2
from multiprocessing import cpu_count


def build_ssh():
    if bool(os.environ.get('SYSTEM_LIBSSH', False)):
        stderr.write("Using system libssh..%s" % (os.sep))
        return
    if os.path.exists('/usr/local/opt/openssl'):
        os.environ['OPENSSL_ROOT_DIR'] = '/usr/local/opt/openssl'

    if not os.path.exists('src'):
        os.mkdir('src')
    if not os.path.exists('local'):
        os.mkdir('local')
    if not os.path.exists('local/lib'):
        os.mkdir('local/lib')
    # Depending on architecture cmake installs libraries into lib64,
    # but we don't care about that.
    if not os.path.exists('local/lib64'):
        os.symlink('lib', 'local/lib64')

    os.chdir('src')
    check_call('cmake -DCMAKE_BUILD_TYPE=Release -DCMAKE_INSTALL_PREFIX=../local -DWITH_NACL=ON -DWITH_PKCS11_URI=ON -DWITH_GSSAPI=1 -DWITH_BLOWFISH_CIPHER=ON ../libssh',
               shell=True, env=os.environ)
    check_call(['make', '-j', 'all', 'install'])
    os.chdir('..')

    for src in glob('local/lib/libssh.so*'):
        copy2(src, 'ssh/')
